/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: CDRInputStream.java,v 1.5 2001/08/13 03:47:33 wood Exp $
 *
 * Date         Author  			Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/
package org.openorb.iiop;

import java.io.*;

import java.util.LinkedList;
import java.util.HashMap;
import java.util.Map;

import org.omg.CORBA.OctetSeqHolder;
import org.omg.CORBA.IntHolder;
import org.omg.CORBA.Any;
import org.omg.CORBA.CompletionStatus;

import org.omg.CORBA.TypeCode;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.StructMember;
import org.omg.CORBA.UnionMember;
import org.omg.CORBA.ValueMember;

import org.omg.SendingContext.RunTime;

import org.openorb.io.*;
import org.openorb.util.Trace;

import org.openorb.util.RepoIDHelper;

/**
 * This class implements CDR for IIOP
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.5 $ $Date: 2001/08/13 03:47:33 $
 */
public class CDRInputStream extends AbstractInputStream
{
	// --------------------------------------------------------------------------
	// Constructors
	// --------------------------------------------------------------------------

	/**
	 * Constructor used by codec. Extending classes must implement a constructor
	 * with this exact signature.
	 */
	public CDRInputStream(org.omg.CORBA.ORB orb, boolean bigEndian, org.omg.GIOP.Version version, StorageBuffer buf) {
		super(buf);

		_orb = orb;

		bigEndian(bigEndian);

		_version = version;

		if(_version.minor == 0) {
			_charEnc = "US-ASCII";
			_wcharEnc = null;
		}
	}

	/**
	 * Constructor used by lower layer. Extending classes must implement a constructor
	 * with this exact signature.
	 */
	public CDRInputStream(org.omg.CORBA.ORB orb, boolean bigEndian, org.omg.GIOP.Version version, BufferSource source) {
		super(source);

		_orb = orb;

		bigEndian(bigEndian);

		_version = version;

		if(_version.minor == 0) {
			_charEnc = "US-ASCII";
			_wcharEnc = null;
		}
	}

	// --------------------------------------------------------------------------
	// Constant for Value Type marshalling / unmarshalling
	// --------------------------------------------------------------------------

	private final static int CODEBASE = 1;
	private final static int NO_TYPE_INFORMATION = 0;
	private final static int SINGLE_TYPE_INFORMATION = 2;
	private final static int MULTIPLE_TYPE_INFORMATION = 6;
	private final static int MASK_TYPE_INFORMATION = 6;
	private final static int CHUNK = 8;

	/**
	 * ORB Reference
	 */
	private org.omg.CORBA.ORB _orb;

	/**
	 * CDR protocol version
	 */
	private org.omg.GIOP.Version _version;

	// non-marked state

	/**
	 * Cache for unmarshalled bits of value. Hashed on index, only one
	 * table is needed as indexes are unique.
	 */
	protected Map _value_cache = new HashMap();

	/**
	 * Bundle of state for mark / reset
	 */
	private MarkState _mark = null;

	/**
	 * Encoding for char data.
	 */
	private String _charEnc = "ISO-8859-1";

	/**
	 * Encoding for wchar data.
	 */
	private String _wcharEnc = "UnicodeBig";

	/**
	 * Alignment for wchar data.
	 */
	private int _wcharAlign = 2;

	/**
	 * reverse wchar data.
	 */
	private boolean _wcharReverse = false;

	// marked state.

	/**
	 * Current index
	 */
	private int _index = 0;

	/**
	 * Flag indicating swap byte order.
	 */
	private boolean _swap = false;

	/**
	 * Remaining bytes in encapsulation
	 */
	private int _encaps_remain = -1;

	/**
	 * Close encapsulation is pending
	 */
	private int _pending_encaps_close = 0;

	/**
	 * remaining size in enclosing encapsulations
	 */
	private LinkedList _encaps_stack = new LinkedList();

	/**
	 * Value nesting level.
	 */
	private int _value_level = 0;

	/**
	 * Continuation chunk for value pending.
	 */
	private boolean _pending_value_reopen = false;

	/**
	 * True if within read for chunked value
	 */
	private boolean _in_chunked_value = false;

	/**
	 * passed between read_value and read_value(java.io.Streamable).
	 * for placement in indirection table.
	 */
	private int _value_indirect = -1;

	/**
	 * count level of continuation chunk. If this is 0 then failure to find
	 * value factory while unmarslaling results in a cancel, otherwise the
	 * exception is thrown without cancelling.
	 */
	private int _continueLevel = 0;

	private class MarkState {
		MarkState() {
			this._index = CDRInputStream.this._index;
			this._swap = CDRInputStream.this._swap;
			this._encaps_remain = CDRInputStream.this._encaps_remain;
			this._pending_encaps_close = CDRInputStream.this._pending_encaps_close;
			this._encaps_stack = new LinkedList(CDRInputStream.this._encaps_stack);
			this._value_level = CDRInputStream.this._value_level;
			this._pending_value_reopen = CDRInputStream.this._pending_value_reopen;
			this._in_chunked_value = CDRInputStream.this._in_chunked_value;
			this._value_indirect = CDRInputStream.this._value_indirect;
		}

		void reset() {
			CDRInputStream.this._index = this._index;
			CDRInputStream.this._swap = this._swap;
			CDRInputStream.this._encaps_remain = this._encaps_remain;
			CDRInputStream.this._pending_encaps_close = this._pending_encaps_close;
			CDRInputStream.this._encaps_stack = this._encaps_stack;
			CDRInputStream.this._value_level = this._value_level;
			CDRInputStream.this._pending_value_reopen = this._pending_value_reopen;
			CDRInputStream.this._in_chunked_value = this._in_chunked_value;
			CDRInputStream.this._value_indirect = this._value_indirect;
		}

		int _index;
		boolean _swap;
		int _encaps_remain = -1;
		int _pending_encaps_close;
		LinkedList _encaps_stack;
		int _value_level;
		boolean _pending_value_reopen;
		boolean _in_chunked_value;
		int _value_indirect;
	}

	/**
	 * tempory vars. here to avoid creating new every time
	 */
	private OctetSeqHolder tmpBuf = new OctetSeqHolder();
	private IntHolder tmpOff = new IntHolder();
	private IntHolder tmpLen = new IntHolder();

	/**
	 * Get the orb associated with the stream.
	 */
	public org.omg.CORBA.ORB orb() {
		return _orb;
	}

	/**
	 * Set the translation codesets.
	 */
	public void setCodesets(int tcsc, int tcsw) {
		if(tcsc != 0) {
			_charEnc = CodeSetDatabase.codesetIDtoEncoding(tcsc);
			if(CodeSetDatabase.codesetIDtoAlignment(tcsc) > 1)
				throw new org.omg.CORBA.CODESET_INCOMPATIBLE(0, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		}

		if(tcsw == 0)
			_wcharEnc = null;
		else {
			_wcharEnc = CodeSetDatabase.codesetIDtoEncoding(tcsw);
			if(_wcharEnc.equals("UnicodeBigUnmarked"))
				_wcharEnc = "UnicodeBig";

			_wcharAlign = CodeSetDatabase.codesetIDtoAlignment(tcsw);
			if(_version.minor == 1 && _wcharAlign != 2)
				throw new org.omg.CORBA.CODESET_INCOMPATIBLE(0, org.omg.CORBA.CompletionStatus.COMPLETED_NO);

      if(_swap && _wcharAlign > 1) {
        try {
          String canonical = CodeSetDatabase.canonicalize(_wcharEnc);
          if(canonical.startsWith("Unicode")) {
            if(_version.minor == 1)
							_wcharEnc = "UnicodeLittle";
          } else
            _wcharReverse = true;
        }
        catch(UnsupportedEncodingException ex) {
          org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
        }
			}
		}
	}

	// --------------------------------------------------------------------------
	// Extended InputStream implementation
	// --------------------------------------------------------------------------

	/**
	 * Get the current index in the buffer.
	 */
	public int index() {
		return _index;
	}

	/**
	 * Reset the index to 0. This affects the alignment. Using this inside an
	 * encapsulation may cause unexpected behaviour.
	 */
	public void reset_index() {
		_index = 0;
	}

	/**
	 * Get the CDR protocol version
	 */
	public org.omg.GIOP.Version version()
	{
		return _version;
	}

	/**
	 * Get the endian mode
	 */
	public boolean bigEndian()
	{
		return !_swap;
	}

	/**
	 * Set the endian mode
	 */
	public void bigEndian( boolean bigEndian )
	{
		_swap = !bigEndian;
	}


    /* Added by Diana on Feb. 15 2002*/
    public byte[] read_arg_buffer()
    {

	byte[] a_b=its_arg_buffer(),arg_buffer=new byte[a_b.length];

	System.arraycopy(a_b,0,arg_buffer,0,a_b.length);
	return arg_buffer;
    }

    /**
     * Adjust alignment.
	 */
	public void alignment(int size)
	{
	    if(_pending_value_reopen) {
		_pending_value_reopen = false;

		int len = peek_long();
		if(!(( len >= 0x7fffff00 ) && ( len <= 0x7fffffff )))
		    begin_value_chunk(read_long());
	    }

	    // Check for alignment
	    if(size > 1) {
		int tmp = _index % size;

		if(tmp != 0)
		    force_skip(size - tmp);
	    }
	}

	/**
	 * Encapsulation begin.
	 */
	public void begin_encapsulation()
	{
		int len = read_ulong();

		_encaps_stack.addLast(new Boolean(_swap));
		_encaps_stack.addLast(new Integer(_encaps_remain));
		_encaps_stack.addLast(new Integer(_index));
		_encaps_stack.addLast(new Boolean(false));

		_encaps_remain = len;
		_in_chunked_value = false;

		_swap = read_boolean();
	}

	/**
	 * skip to end of current encapsulation
	 */
	public void end_encapsulation()
	{
		if(_pending_encaps_close == 0)
			force_skip(_encaps_remain);
		--_pending_encaps_close;
	}

	// --------------------------------------------------------------------------
	// java.io.InputStream implementation
	// --------------------------------------------------------------------------

	/**
	 * Skip over bytes.
	 */
	public long skip(long count)
	{
		long total = super.skip(count);
		postread((int)total);
		return total;
	}

	public void mark(int readlimit) {
		super.mark(readlimit);
		_mark = new MarkState();
	}

	public void reset()
		throws IOException
	{
		super.reset();
		_mark.reset();
		_mark = null;
	}

	// --------------------------------------------------------------------------
	// Standard operations
	// --------------------------------------------------------------------------

	/**
	 * Return an IDL boolean
	 */
	public boolean read_boolean()
	{
		return (read_octet() == 1);
	}

	/**
	 * Return an IDL char
	 */
	public char read_char()
	{
		alignment( 1 );

		tmpLen.value = 1;
		next(tmpBuf, tmpOff, tmpLen);
		postread(1);

		try {
			String s = new String(tmpBuf.value, tmpOff.value, 1, _charEnc);

			if(s.length() == 1)
				return s.charAt(0);

			cancel(new org.omg.CORBA.MARSHAL("Unable to decode char value",
				IIOPMinorCodes.MARSHAL_CHAR, CompletionStatus.COMPLETED_MAYBE));
		}
		catch(UnsupportedEncodingException ex) {
			org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
		}
		return '\0';
	}

	/**
	 * Return an IDL wchar
	 */
	public char read_wchar()
	{
		if(_wcharEnc == null) {
			if(_version.minor == 0)
				cancel(new org.omg.CORBA.BAD_OPERATION("Wchar not supported in IIOP 1.0",
					IIOPMinorCodes.BAD_OPERATION_IIOP_VERSION, CompletionStatus.COMPLETED_MAYBE));
			else
				cancel(new org.omg.CORBA.MARSHAL("Missing wchar encoder.",
					IIOPMinorCodes.MARSHAL_WCHAR, CompletionStatus.COMPLETED_MAYBE));
			return '\0';
		}

		byte [] buf = null;
		int off = 0;
		int len = 0;
		switch(_version.minor) {
		case 0:
			cancel(new org.omg.CORBA.BAD_OPERATION("Wchar not supported in IIOP 1.0",
				IIOPMinorCodes.BAD_OPERATION_IIOP_VERSION, CompletionStatus.COMPLETED_MAYBE));
			return '\0';
		case 1:
			alignment(2);
			len = tmpLen.value = 2;
			next(tmpBuf, tmpOff, tmpLen);

			if(tmpLen.value != 0) {
				buf = new byte[2];
				if(_wcharReverse) {
					buf[1] = tmpBuf.value[tmpOff.value];
					next(tmpBuf, tmpOff, tmpLen);
					buf[0] = tmpBuf.value[tmpOff.value];
				} else {
					buf[0] = tmpBuf.value[tmpOff.value];
					next(tmpBuf, tmpOff, tmpLen);
					buf[1] = tmpBuf.value[tmpOff.value];
				}
				off = 0;
			} else if(_wcharReverse) {
				buf = new byte[2];
				buf[0] = tmpBuf.value[1];
				buf[1] = tmpBuf.value[0];
				off = 0;
			} else {
				buf = tmpBuf.value;
				off = tmpOff.value;
			}

			postread(2);
			break;
		case 2:
			alignment(1);
			tmpLen.value = 1;
			next(tmpBuf, tmpOff, tmpLen);

			len = tmpLen.value = tmpBuf.value[tmpOff.value];
			next(tmpBuf, tmpOff, tmpLen);

			postread(len-tmpLen.value+1);
			if(tmpLen.value != 0) {
				buf = new byte[len];
				off = 0;
				System.arraycopy(tmpBuf.value, tmpOff.value, buf, 0, len-tmpLen.value);
				read_octet_array(buf, len-tmpLen.value, tmpLen.value);
				if(_wcharReverse) {
					byte tmp;
					for(int i = 0; i < _wcharAlign/2; ++i) {
						tmp = buf[i];
						buf[i] = buf[_wcharAlign-i];
						buf[_wcharAlign-i] = tmp;
					}
				}
			} else if(_wcharReverse) {
				buf = new byte[len];
				off = 0;
				for(int i = 0; i < _wcharAlign; ++i)
					buf[i] = tmpBuf.value[tmpOff.value+_wcharAlign-i];
			} else {
				buf = tmpBuf.value;
				off = tmpOff.value;
			}
			break;
		}

		try {
			String s = new String(buf, off, len, _wcharEnc);

			if(s.length() == 1)
				return s.charAt(0);

			cancel(new org.omg.CORBA.MARSHAL("Unable to decode wchar value",
				IIOPMinorCodes.MARSHAL_WCHAR, CompletionStatus.COMPLETED_MAYBE));
		}
		catch(UnsupportedEncodingException ex) {
			org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
		}
		return '\0';
	}

	/**
	 * Return an IDL octet
	 */
	public byte read_octet()
	{
		alignment( 1 );

		tmpLen.value = 1;
		next(tmpBuf, tmpOff, tmpLen);

		byte b = tmpBuf.value[tmpOff.value];

		postread(1);

		return b;
	}

	/**
	 * Return an IDL short
	 */
	public short read_short()
	{
		alignment(2);

		int val;

		tmpLen.value = 2;
		if(next(tmpBuf, tmpOff, tmpLen) == 2) {
			val = ((tmpBuf.value[tmpOff.value  ]&0xFF) << ( _swap ? 0 : 8 )) |
				  ((tmpBuf.value[tmpOff.value+1]&0xFF) << ( _swap ? 8 : 0 ));
		} else {
			val =  ((tmpBuf.value[tmpOff.value]&0xFF) << ( _swap ? 0 : 8 ));
			next(tmpBuf, tmpOff, tmpLen);
			val = val | ((tmpBuf.value[tmpOff.value]&0xFF) << ( _swap ? 8 : 0 ));
		}

		postread(2);

		return ( short ) val ;
	}

	/**
	 * Return an IDL unsigned short
	 */
	public short read_ushort()
	{
		return read_short() ;
	}

	/**
	 * Return an IDL long
	 */
	public int read_long()
	{
		alignment(4);

		int val;

		int got;
		tmpLen.value = 4;
		if(( got = next(tmpBuf, tmpOff, tmpLen)) == 4) {
			// optimize for normal case.
			val = ((tmpBuf.value[tmpOff.value  ]&0xFF) << ( _swap ? 0 : 24 )) |
				  ((tmpBuf.value[tmpOff.value+1]&0xFF) << ( _swap ? 8 : 16 )) |
				  ((tmpBuf.value[tmpOff.value+2]&0xFF) << ( _swap ? 16 : 8 )) |
				  ((tmpBuf.value[tmpOff.value+3]&0xFF) << ( _swap ? 24 : 0 ));
		} else {
			// general case. This will probably never occour.
			val = 0;
			int shf = _swap?0:24;
			while(true) {
				for(int i = 0; i < got; ++i) {
					val = val | ((tmpBuf.value[tmpOff.value+i]&0xFF) << shf);
					shf += _swap?8:-8;
				}
				if(_swap?(shf < 24):(shf > 0)) {
					got = next(tmpBuf, tmpOff, tmpLen);
					continue;
				}
				break;
			}
		}

		postread(4);

		return val ;
	}

  private int peek_long() {
    mark(4);
    int ret = read_long();
    try {
  		reset();
    }
    catch(java.io.IOException ex) {
      org.openorb.util.Trace.assertTrue(false, ex.toString());
    }
    return ret;
  }

	/**
	 * Return an IDL unsigned long
	 */
	public int read_ulong()
	{
		return read_long() ;
	}

	/**
	 * Return an IDL long long
	 */
	public long read_longlong()
	{
		alignment(8);

		long val;

		int got;
		tmpLen.value = 8;
		if(( got = next(tmpBuf, tmpOff, tmpLen)) == 8) {
			// optimize for normal case.
			val = ((tmpBuf.value[tmpOff.value  ]&0xFFl) << ( _swap ? 0l : 56l )) |
				  ((tmpBuf.value[tmpOff.value+1]&0xFFl) << ( _swap ? 8l : 48l )) |
				  ((tmpBuf.value[tmpOff.value+2]&0xFFl) << ( _swap ? 16l : 40l )) |
				  ((tmpBuf.value[tmpOff.value+3]&0xFFl) << ( _swap ? 24l : 32l )) |
				  ((tmpBuf.value[tmpOff.value+4]&0xFFl) << ( _swap ? 32l : 24l )) |
				  ((tmpBuf.value[tmpOff.value+5]&0xFFl) << ( _swap ? 40l : 16l )) |
				  ((tmpBuf.value[tmpOff.value+6]&0xFFl) << ( _swap ? 48l : 8l )) |
				  ((tmpBuf.value[tmpOff.value+7]&0xFFl) << ( _swap ? 56l : 0l ));
		} else {
			// general case. This will probably never occour.
			val = 0;
			long shf = _swap?0:56;
			while(true) {
				for(int i = 0; i < got; ++i) {
					val = val | ((long)(tmpBuf.value[tmpOff.value+i]&0xFF) << shf);
					shf += _swap?8:-8;
				}
				if(_swap?(shf < 56):(shf > 0)) {
					got = next(tmpBuf, tmpOff, tmpLen);
					continue;
				}
				break;
			}
		}

		postread(8);

		return val;
	}

	/**
	 * Return an IDL unsigned long long
	 */
	public long read_ulonglong()
	{
		return read_longlong() ;
	}

	/**
	 * Return an IDL float
	 */
	public float read_float()
	{
		return Float.intBitsToFloat(read_long());
	}

	/**
	 * Return an IDL double
	 */
	public double read_double()
	{
		return Double.longBitsToDouble(read_longlong()) ;
	}

	/**
	 * Return an IDL string
	 */
	public String read_string()
	{
		int len = read_ulong();

		byte [] buf;
		int off;

		tmpLen.value = len;
		next(tmpBuf, tmpOff, tmpLen);

		postread(len-tmpLen.value);
		if(tmpLen.value == 0) {
			buf = tmpBuf.value;
			off = tmpOff.value;
		} else {
			buf = new byte[len];
			off = 0;
			System.arraycopy(tmpBuf.value, tmpOff.value, buf, 0, len-tmpLen.value);
			read_octet_array(buf, len-tmpLen.value, tmpLen.value);
		}

		// ignore terminating null
		try {
			return new String(buf, off, len-1, _charEnc);
		}
		catch(UnsupportedEncodingException ex) {
			org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
		}
		return null;
	}

	/**
	 * Return an IDL wstring
	 */
	public String read_wstring()
	{
		if(_wcharEnc == null) {
			if(_version.minor == 0)
				cancel(new org.omg.CORBA.BAD_OPERATION("Wchar not supported in IIOP 1.0",
					IIOPMinorCodes.BAD_OPERATION_IIOP_VERSION, CompletionStatus.COMPLETED_MAYBE));
			else
				cancel(new org.omg.CORBA.MARSHAL("Missing wchar encoder.",
					IIOPMinorCodes.MARSHAL_WCHAR, CompletionStatus.COMPLETED_MAYBE));
			return null;
		}

		if(_version.minor == 0) {
			cancel(new org.omg.CORBA.BAD_OPERATION("Wchar not supported in IIOP 1.0",
				IIOPMinorCodes.BAD_OPERATION_IIOP_VERSION, CompletionStatus.COMPLETED_MAYBE));
			return null;
		}

		int len = read_ulong();

		if(_version.minor == 1 && _wcharAlign > 1)
			len *= _wcharAlign;

		byte [] buf;
		int off;

		tmpLen.value = len;
		next(tmpBuf, tmpOff, tmpLen);

		postread(len-tmpLen.value);
		if(tmpLen.value != 0) {
			buf = new byte[len];
			off = 0;
			System.arraycopy(tmpBuf.value, tmpOff.value, buf, 0, len-tmpLen.value);
			read_octet_array(buf, len-tmpLen.value, tmpLen.value);
			if(_wcharReverse) {
				byte tmp;
				for(int i = 0; i < len; i+=_wcharAlign)
					for(int j = 0; j < _wcharAlign/2; ++j) {
						tmp = buf[off+i+j];
						buf[off+i+j] = buf[off+i+_wcharAlign-j];
						buf[off+i+_wcharAlign-j] = tmp;
					}
			}
		} else if(_wcharReverse) {
			buf = new byte[len];
			off = 0;
			for(int i = 0; i < len; i+=_wcharAlign)
				for(int j = 0; j < _wcharAlign; ++j)
					buf[i+j] = tmpBuf.value[tmpOff.value+i+_wcharAlign-j];
		} else {
			buf = tmpBuf.value;
			off = tmpOff.value;
		}

		if(_version.minor == 1)
			len -= (_wcharAlign==0)?1:_wcharAlign;

		if(len == 0)
			return "";

		try {
			return new String(buf, off, len, _wcharEnc);
		}
		catch(UnsupportedEncodingException ex) {
			org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
		}
		return null;
	}

	/**
	 * Read an IDL boolean array
	 */
	public void read_boolean_array(boolean[] val, int off, int len)
	{
		alignment(1);

		int rd;
		tmpLen.value = len;
		while(tmpLen.value > 0) {
			rd = next(tmpBuf, tmpOff, tmpLen);
			for(int i = 0; i < rd; ++i)
				val[off+i] = (tmpBuf.value[tmpOff.value+i] == 1);
			off += rd;
		}

		postread(len);
	}

	/**
	 * Read an IDL char array from the stream
	 */
	public void read_char_array(char[] val, int voff, int len)
	{
		alignment(1);

		byte [] buf;
		int off;

		tmpLen.value = len;
		next(tmpBuf, tmpOff, tmpLen);

		postread(len-tmpLen.value);
		if(tmpLen.value == 0) {
			buf = tmpBuf.value;
			off = tmpOff.value;
		} else {
			buf = new byte[len];
			off = 0;
			System.arraycopy(tmpBuf.value, tmpOff.value, buf, 0, len-tmpLen.value);
			read_octet_array(buf, len-tmpLen.value, tmpLen.value);
		}

		try {
			String s = new String(buf, off, len, _charEnc);

			if(s.length() != len)
				cancel(new org.omg.CORBA.MARSHAL("Unable to decode char value",
					IIOPMinorCodes.MARSHAL_CHAR, CompletionStatus.COMPLETED_MAYBE));

			s.getChars(0, len, val, voff);
		}
		catch(UnsupportedEncodingException ex) {
			org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
		}
	}

	/**
	 * Read a wchar array
	 */
	public void read_wchar_array(char[] val, int voff, int vlen)
	{
		if(_wcharEnc == null) {
			if(_version.minor == 0)
				cancel(new org.omg.CORBA.BAD_OPERATION("Wchar not supported in IIOP 1.0",
					IIOPMinorCodes.BAD_OPERATION_IIOP_VERSION, CompletionStatus.COMPLETED_MAYBE));
			else
				cancel(new org.omg.CORBA.MARSHAL("Missing wchar encoder.",
					IIOPMinorCodes.MARSHAL_WCHAR, CompletionStatus.COMPLETED_MAYBE));
			return;
		}

		byte [] buf;
		int off;
		int len;
		switch(_version.minor) {
		case 0:
			cancel(new org.omg.CORBA.MARSHAL("Wchar not supported in IIOP 1.0",
				IIOPMinorCodes.BAD_OPERATION_IIOP_VERSION, CompletionStatus.COMPLETED_MAYBE));
			break;
		case 1:
			alignment(2);
			len = tmpLen.value = 2*vlen;
			next(tmpBuf, tmpOff, tmpLen);

			postread(len-tmpLen.value);
			if(tmpLen.value != 0) {
				buf = new byte[len];
				off = 0;
				System.arraycopy(tmpBuf.value, tmpOff.value, buf, 0, len-tmpLen.value);
				read_octet_array(buf, len-tmpLen.value, tmpLen.value);
				if(_wcharReverse) {
					byte tmp;
					for(int i = 0; i < len; i += 2) {
						tmp = buf[i];
						buf[i] = buf[i+1];
						buf[i+1] = tmp;
					}
				}
			} else if(_wcharReverse) {
				buf = new byte[len];
				off = 0;
				for(int i = 0; i < len; i+=2) {
					buf[0] = tmpBuf.value[tmpOff.value+i+1];
					buf[1] = tmpBuf.value[tmpOff.value+i];
				}
			} else {
				buf = tmpBuf.value;
				off = tmpOff.value;
			}

			try {
				String s = new String(buf, off, len, _wcharEnc);

				if(s.length() != vlen)
					cancel(new org.omg.CORBA.MARSHAL("Unable to decode char value",
						IIOPMinorCodes.MARSHAL_CHAR, CompletionStatus.COMPLETED_MAYBE));
				s.getChars(0, vlen, val, voff);
			}
			catch(UnsupportedEncodingException ex) {
				org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
			}
			return;
		case 2:
			alignment(1);
			int post = 0;
			for(int c = 0; c < vlen; ++c) {
				tmpLen.value = 1;
				next(tmpBuf, tmpOff, tmpLen);

				len = tmpLen.value = tmpBuf.value[tmpOff.value];
				next(tmpBuf, tmpOff, tmpLen);

				if(tmpLen.value != 0) {
					postread(post+len+1-tmpLen.value);
					post = 0;

					buf = new byte[len];
					off = 0;
					System.arraycopy(tmpBuf.value, tmpOff.value, buf, 0, len-tmpLen.value);
					read_octet_array(buf, len-tmpLen.value, tmpLen.value);

					if(_wcharReverse) {
						byte tmp;
						for(int i = 0; i < _wcharAlign/2; ++i) {
							tmp = buf[off+i];
							buf[off+i] = buf[off+_wcharAlign-i];
							buf[off+_wcharAlign-i] = tmp;
						}
					}
				} else if(_wcharReverse) {
					buf = new byte[len];
					off = 0;
					for(int i = 0; i < _wcharAlign; ++i)
						buf[i] = tmpBuf.value[tmpOff.value+_wcharAlign-1];
				} else {
					post += len+1;
					buf = tmpBuf.value;
					off = tmpOff.value;
				}

				try {
					String s = new String(buf, off, len, _wcharEnc);

					if(s.length() != 1)
						cancel(new org.omg.CORBA.MARSHAL("Unable to decode wchar value",
							IIOPMinorCodes.MARSHAL_WCHAR, CompletionStatus.COMPLETED_MAYBE));
					val[voff+c] = s.charAt(0);
				}
				catch(UnsupportedEncodingException ex) {
					org.openorb.util.Trace.assertTrue(false, "Unsupported encoding should be impossible");
				}
			}
			if(post > 0)
				postread(post);
			break;
		}
	}

	/**
	 * Read an octet array
	 */
	public void read_octet_array(byte[] val, int off, int len)
	{
		alignment(1);

		int rd;
		tmpLen.value = len;
		while(tmpLen.value > 0) {
			rd = next(tmpBuf, tmpOff, tmpLen);
			System.arraycopy(tmpBuf.value, tmpOff.value, val, off, rd);
			off += rd;
		}

		postread(len);
	}

	/**
	 * Read a short array
	 */
	public void read_short_array(short[] val, int off, int len)
	{
		alignment(2);

		tmpLen.value = len*2;

		int rd;
		while(tmpLen.value > 0) {
			rd = next(tmpBuf, tmpOff, tmpLen);
			for(int i = 0; i < rd-1; i += 2, ++off)
				val[off] = (short)(((tmpBuf.value[tmpOff.value+i]&0xFF) << (_swap?0:8)) |
								   ((tmpBuf.value[tmpOff.value+i+1]&0xFF) << (_swap?8:0)));

			if(rd%2 != 0) {
				int hlf = ((tmpBuf.value[tmpOff.value+rd]&0xFF) << (_swap?0:8));
				int tmp = tmpLen.value;
				tmpLen.value = 1;

				next(tmpBuf, tmpOff, tmpLen);
				val[off] = (short)(hlf | ((tmpBuf.value[tmpOff.value]&0xFF) << (_swap?8:0)));

				++off;
				tmpLen.value = tmp-1;
			}
		}

		postread(len*2);
	}

	/**
	 * Read an unsigned short array
	 */
	public void read_ushort_array(short[] val, int off, int len)
	{
		read_short_array(val, off, len);
	}

	/**
	 * Read a long array
	 */
	public void read_long_array(int[] val, int off, int len)
	{
		alignment(4);

		tmpLen.value = len*4;

		int rd;
		while(tmpLen.value > 0) {
			rd = next(tmpBuf, tmpOff, tmpLen);
			for(int i = 0; i < rd-3; i += 4, ++off)
				val[off] = (((tmpBuf.value[tmpOff.value+i  ]&0xFF) << (_swap?0:24)) |
							((tmpBuf.value[tmpOff.value+i+1]&0xFF) << (_swap?8:16)) |
							((tmpBuf.value[tmpOff.value+i+2]&0xFF) << (_swap?16:8)) |
							((tmpBuf.value[tmpOff.value+i+3]&0xFF) << (_swap?24:0)));

			// handle unusual cases
			if(rd%4 != 0) {
				val[off] = 0;

				int rcvd = rd%4;
				rd -= rd%4;

				int s = 0;
				for(; s < rcvd; ++s)
					val[off] = val[off] | ((tmpBuf.value[tmpOff.value+rd+s]&0xFF)
										   << (_swap?(8*s):(24-8*s)));

				int tmp = tmpLen.value - (4-s);
				while(s<4) {
					tmpLen.value = 4-s;
					rd = next(tmpBuf, tmpOff, tmpLen);
					for(int i = 0; i < rd; ++i, ++s)
						val[off] = val[off] | ((tmpBuf.value[tmpOff.value+i]&0xFF)
											   << (_swap?(8*s):(24-8*s)));
				}

				tmpLen.value = tmp;
			}
		}

		postread(len*4);
	}

	/**
	 * Read an unsigned long array
	 */
	public void read_ulong_array(int[] val, int off, int len)
	{
		read_long_array(val, off, len);
	}

	/**
	 * Read a long long array
	 */
	public void read_longlong_array(long[] val, int off, int len)
	{
		alignment(8);

		tmpLen.value = len*8;

		int rd;
		while(tmpLen.value > 0) {
			rd = next(tmpBuf, tmpOff, tmpLen);
			for(int i = 0; i < rd-7; i += 8, ++off)
				val[off] = (((long)(tmpBuf.value[tmpOff.value+i  ]&0xFF) << (long)(_swap?0:56)) |
							((long)(tmpBuf.value[tmpOff.value+i+1]&0xFF) << (long)(_swap?8:48)) |
							((long)(tmpBuf.value[tmpOff.value+i+2]&0xFF) << (long)(_swap?16:40)) |
							((long)(tmpBuf.value[tmpOff.value+i+3]&0xFF) << (long)(_swap?24:32)) |
							((long)(tmpBuf.value[tmpOff.value+i+4]&0xFF) << (long)(_swap?32:24)) |
							((long)(tmpBuf.value[tmpOff.value+i+5]&0xFF) << (long)(_swap?40:16)) |
							((long)(tmpBuf.value[tmpOff.value+i+6]&0xFF) << (long)(_swap?48:8)) |
							((long)(tmpBuf.value[tmpOff.value+i+7]&0xFF) << (long)(_swap?56:0)));

			// handle unusual cases
			if(rd%8 != 0) {
				val[off] = 0l;

				int rcvd = rd%8;
				rd -= rd%8;

				int s = 0;
				for(; s < rcvd; ++s)
					val[off] = val[off] | ((long)(tmpBuf.value[tmpOff.value+rd+s]&0xFF)
										   << (long)(_swap?(8*s):(56-8*s)));

				int tmp = tmpLen.value - (8-s);
				while(s<8) {
					tmpLen.value = 8-s;
					rd = next(tmpBuf, tmpOff, tmpLen);
					for(int i = 0; i < rd; ++i, ++s)
						val[off] = val[off] | ((long)(tmpBuf.value[tmpOff.value+i]&0xFF)
											   << (long)(_swap?(8*s):(56-8*s)));
				}

				tmpLen.value = tmp;
			}
		}

		postread(len*8);
	}

	/**
	 * Read an unsigned long long array
	 */
	public void read_ulonglong_array(long[] val, int off, int len)
	{
		read_longlong_array(val, off, len);
	}

	/**
	 * Read a float array
	 */
	public void read_float_array(float[] val, int off, int len)
	{
		int [] vals = new int[len];
		read_long_array(vals, 0, len);

		for(int i = 0; i < len; ++i)
			val[off+i] = Float.intBitsToFloat(vals[i]);
	}

	/**
	 * Read a double array
	 */
	public void read_double_array(double[] val, int off, int len)
	{
		long [] vals = new long[len];
		read_longlong_array(vals, 0, len);

		for(int i = 0; i < len; ++i)
			val[off+i] = Double.longBitsToDouble(vals[i]);
	}

	/**
	 * Read a typecode
	 */
	public org.omg.CORBA.TypeCode read_TypeCode()
	{
		return read_TypeCodeValue(new HashMap());
	}

	private org.omg.CORBA.TypeCode read_TypeCodeValue(HashMap cache)
	{
		TypeCode typeco = null;

		alignment(4);
		Integer kind_index = new Integer(_index);
		int kind = read_ulong();

		// scratch vars.
		String id , name;
		TypeCode contain;
		int length;

		switch(kind)
		{
		case org.omg.CORBA.TCKind._tk_null :
		case org.omg.CORBA.TCKind._tk_void :
		case org.omg.CORBA.TCKind._tk_short :
		case org.omg.CORBA.TCKind._tk_long :
		case org.omg.CORBA.TCKind._tk_ushort :
		case org.omg.CORBA.TCKind._tk_ulong :
		case org.omg.CORBA.TCKind._tk_float :
		case org.omg.CORBA.TCKind._tk_double :
		case org.omg.CORBA.TCKind._tk_boolean :
		case org.omg.CORBA.TCKind._tk_char :
		case org.omg.CORBA.TCKind._tk_octet :
		case org.omg.CORBA.TCKind._tk_any :
		case org.omg.CORBA.TCKind._tk_TypeCode :
		case org.omg.CORBA.TCKind._tk_Principal :

		case org.omg.CORBA.TCKind._tk_longlong:
		case org.omg.CORBA.TCKind._tk_ulonglong:
		case org.omg.CORBA.TCKind._tk_longdouble:
		case org.omg.CORBA.TCKind._tk_wchar:
			typeco = _orb.get_primitive_tc(TCKind.from_int(kind));
			break;
		case org.omg.CORBA.TCKind._tk_objref:
			begin_encapsulation();
			id = read_string();
			name = read_string();
			end_encapsulation();

			typeco = _orb.create_interface_tc(id,name);
			break;
		case org.omg.CORBA.TCKind._tk_struct :
		{
			begin_encapsulation();
			id = read_string();
			name = read_string();
			cache.put(kind_index, id);

			length = read_ulong();
			StructMember[] sm = new StructMember[length];
			for (int i = 0 ; i < length; i++)
			{
				sm[i] = new StructMember();
				sm[i].name = read_string();
				sm[i].type = read_TypeCodeValue(cache);
			}
			end_encapsulation();

			typeco = _orb.create_struct_tc(id,name,sm);
		}	break;
		case org.omg.CORBA.TCKind._tk_union :
		{
			begin_encapsulation();
			id = read_string();
			name = read_string();
			cache.put(kind_index, id);

			TypeCode disc = read_TypeCodeValue(cache);
			int defaultused = read_long();
			length = read_ulong() ;
			UnionMember[] um = new UnionMember[length];
			for(int i = 0 ; i< length; i++)
			{
				um[i] = new UnionMember();

				um[i].label = _orb.create_any();
				um[i].label.read_value( this, disc );
				if ( i == defaultused )
					um[i].label.insert_octet((byte)0);

				um[i].name = read_string();
				um[i].type = read_TypeCodeValue(cache);
			}
			end_encapsulation();

			typeco = _orb.create_union_tc(id,name,disc,um);
		} break;
		case org.omg.CORBA.TCKind._tk_enum :
		{
			begin_encapsulation();
			id = read_string();
			name = read_string();

			length = read_ulong();
			String [] members = new String[length];
			for(int i = 0 ; i < length ; i++)
				members[i] = read_string();

			end_encapsulation();

			typeco = _orb.create_enum_tc(id,name,members);
		}	break;
		case org.omg.CORBA.TCKind._tk_string :
			length = read_ulong();
			typeco = _orb.create_string_tc(length);
			break;
		case org.omg.CORBA.TCKind._tk_sequence :
			begin_encapsulation();
			contain = read_TypeCodeValue(cache);
			length = read_ulong();
			end_encapsulation();

			typeco = _orb.create_sequence_tc(length,contain);
			break;
		case org.omg.CORBA.TCKind._tk_array :
			begin_encapsulation();
			contain = read_TypeCodeValue(cache);
			length = read_ulong();
			end_encapsulation();

			typeco = _orb.create_array_tc(length,contain);
			break;
		case org.omg.CORBA.TCKind._tk_alias :
			begin_encapsulation();
			id = read_string();
			name = read_string();
			cache.put(kind_index, id);

			contain = read_TypeCodeValue(cache);
			end_encapsulation();

			typeco = _orb.create_alias_tc(id,name,contain);
			break;
		case org.omg.CORBA.TCKind._tk_except:
		{
			begin_encapsulation();
			id = read_string();
			name = read_string();
			cache.put(kind_index, id);

			length = read_ulong();
			StructMember[] sm = new StructMember[length];
			for(int i = 0 ; i < length; i++)
			{
				sm[i] = new StructMember();
				sm[i].name = read_string();
				sm[i].type = read_TypeCodeValue(cache);
			}
			end_encapsulation();

			typeco = _orb.create_exception_tc(id,name,sm);

			cache.put(kind_index, typeco);
		} break;
		case org.omg.CORBA.TCKind._tk_wstring:
			length = read_ulong();
			typeco = _orb.create_wstring_tc(length);
			break;
		case org.omg.CORBA.TCKind._tk_fixed:
		{
			short digits = read_ushort();
			short scale = read_ushort();
			typeco = _orb.create_fixed_tc(digits, scale);
		}	break;
		case org.omg.CORBA.TCKind._tk_value :
		{
			begin_encapsulation();
			id = read_string();
			name = read_string();
			cache.put(kind_index, id);

			short modifier = read_short();
			contain = read_TypeCodeValue(cache);
			length = read_ulong();
			ValueMember[] vm = new ValueMember[length];
			for ( int i=0; i<length; i++ )
			{
				vm[i] = new ValueMember();
				vm[i].name = read_string();
				vm[i].type = read_TypeCodeValue(cache);
				vm[i].access = read_short();
			}
			end_encapsulation();

			typeco = _orb.create_value_tc(id,name,modifier,contain,vm);
		}	break;
		case org.omg.CORBA.TCKind._tk_value_box :
			begin_encapsulation();
			id = read_string();
			name = read_string();
			cache.put(kind_index, id);

			contain = read_TypeCodeValue(cache);
			end_encapsulation();

			typeco = _orb.create_value_box_tc(id,name,contain);
			break;
		case org.omg.CORBA.TCKind._tk_native:
			begin_encapsulation();
			id = read_string();
			name = read_string();
			end_encapsulation();

			typeco = _orb.create_native_tc(id,name);
			break;
		case org.omg.CORBA.TCKind._tk_abstract_interface:
			begin_encapsulation();
			id = read_string();
			name = read_string();
			end_encapsulation();

			typeco = _orb.create_abstract_interface_tc(id,name);
			break;
		/*
		case org.omg.CORBA.TCKind._tk_home:
			begin_encapsulation();
			id = read_string();
			name = read_string();
			end_encapsulation();

			typeco = _orb.create_home_tc(id,name);
			break;
	 	*/
		/*
		case org.omg.CORBA.TCKind._tk_component:
			begin_encapsulation();
			id = read_string();
			name = read_string();
			end_encapsulation();

			typeco = _orb.create_component_tc(id,name);
			break;
		*/
		case 0xffffffff :
		{
			int oldidx = _index + read_long();

			Object lookup = cache.get(new Integer(oldidx));
			// either a recursive typecode or a repeated typecode.
			if(lookup instanceof String)
				typeco = _orb.create_recursive_tc((String)lookup);
			else if(lookup != null)
				typeco = (TypeCode)lookup;
			else
				cancel(new org.omg.CORBA.MARSHAL("Invalid typecode offset",
					IIOPMinorCodes.MARSHAL_TC_OFFSET, CompletionStatus.COMPLETED_MAYBE));
		} break;
		default:
			cancel(new org.omg.CORBA.MARSHAL("Invalid typecode kind",
				IIOPMinorCodes.MARSHAL_TC_KIND, CompletionStatus.COMPLETED_MAYBE));
		}
		cache.put(kind_index, typeco);

		return typeco;
	}

	/**
	 * Read an any
	 */
	public Any read_any()
	{
		TypeCode typeco = read_TypeCode();
		org.omg.CORBA.Any any = _orb.create_any();
		any.read_value(this,typeco);
		return any;
	}

	/**
	 * Read a principal
	 */
	public org.omg.CORBA.Principal read_Principal()
	{
		byte[] b = new byte[read_ulong()];
		read_octet_array(b, 0 ,b.length);
		return new org.openorb.CORBA.Principal(b);
	}

	/**
	 * Read an object
	 */
	public org.omg.CORBA.Object read_Object()
	{
		org.omg.IOP.IOR ior = org.omg.IOP.IORHelper.read(this);

		if(ior.type_id.length() == 0 && ior.profiles.length == 0)
			return null;

		return new org.openorb.CORBA.ObjectStub(_orb, ior);
	}

	/**
	 * Read an Object
	 */
	public org.omg.CORBA.Object read_Object( Class clz )
	{
		org.omg.IOP.IOR ior = org.omg.IOP.IORHelper.read(this);
		if(ior.type_id.length() == 0 && ior.profiles.length == 0)
			return null;

		if(clz == null || !org.omg.CORBA.Object.class.isAssignableFrom(clz))
			return new org.openorb.CORBA.ObjectStub(_orb, ior);

		org.omg.CORBA.portable.ObjectImpl ret = null;
    try {
      ret = (org.omg.CORBA.portable.ObjectImpl)clz.newInstance();
    }
    catch(Exception ex) {
      cancel(new org.omg.CORBA.BAD_PARAM("Unable to instantiate class",
        IIOPMinorCodes.BAD_PARAM_OBJ_CLASS, CompletionStatus.COMPLETED_NO));
    }

		org.omg.CORBA.portable.Delegate ndeleg
			= new org.openorb.CORBA.Delegate(_orb, ior);

		ret._set_delegate(ndeleg);

		return ret;
	}

	/**
	 * Read a context
	 */
	public org.omg.CORBA.Context read_Context()
	{
		org.omg.CORBA.NVList nv = _orb.create_list(0);

		int max = ( read_ulong() / 2 );
		for ( int i=0; i<max; i++ )
		{
			org.omg.CORBA.Any a = _orb.create_any();
			String name = read_string();
			a.insert_string(read_string());
			nv.add_value(name,a,0);
		}

		org.omg.CORBA.Context context = new org.openorb.CORBA.dii.Context("",null,_orb);
		context.set_values(nv);

		return context;
	}

	/**
	 * Read a fixed
	 *
	 * @deprecated Loses scale and precision, see
	 *              http://www.omg.org/issues/issue3431.txt
	 */
	public java.math.BigDecimal read_fixed()
	{
		return read_fixed((short)-1, (short)-1);
	}

	/**
	 * read a fixed.
	 */
	public java.math.BigDecimal read_fixed(org.omg.CORBA.TypeCode type) {
		try {
			return read_fixed(type.fixed_digits(), type.fixed_scale());
		}
		catch(org.omg.CORBA.TypeCodePackage.BadKind ex) {
			cancel(new org.omg.CORBA.BAD_PARAM("Typecode is not fixed type",
				IIOPMinorCodes.BAD_PARAM_FIXED_TYPE, CompletionStatus.COMPLETED_NO));
			return null;
		}
	}

	/**
	 * read a fixed.
	 */
	public java.math.BigDecimal read_fixed(short digits, short scale) {
		StringBuffer value = new StringBuffer();

		int count = 0;

		boolean body = false;
		boolean negative = false;
		while(true) {
			byte v = read_octet();

			int hi = (v >>> 4) & 0xF;
			int lo = v & 0xF;

			if(hi > 9)
				cancel(new org.omg.CORBA.MARSHAL("Error in fixed format",
					IIOPMinorCodes.MARSHAL_FIXED, CompletionStatus.COMPLETED_MAYBE));

			if(body || hi > 0) {
				value.append('0'+hi);
				count++;
			}

			if(lo < 0xA) {
				value.append('0'+lo);
				count++;
			}	else if(lo == 0xC) {
				negative = true;
				break;
			} else if (lo == 0xD)
				break;
			else
				cancel(new org.omg.CORBA.MARSHAL("Error in fixed format",
					IIOPMinorCodes.MARSHAL_FIXED, CompletionStatus.COMPLETED_MAYBE));

			body = true;
		}

		if(digits >= 0 && count > digits)
			cancel(new org.omg.CORBA.MARSHAL("Fixed data does not match type",
				IIOPMinorCodes.MARSHAL_FIXED, CompletionStatus.COMPLETED_MAYBE));

		java.math.BigDecimal ret;
		try
		{
			ret = new java.math.BigDecimal(value.toString());
		}
		catch(NumberFormatException ex)
		{
			cancel(new org.omg.CORBA.MARSHAL("Error in fixed format",
				IIOPMinorCodes.MARSHAL_FIXED, CompletionStatus.COMPLETED_MAYBE));
			return null;
		}

		if(negative)
			ret = ret.negate();

		if(scale > 0)
			ret = ret.movePointLeft(scale);

		return ret;
	}

	/**
	 * Read a value from a CDR stream
	 */
	public java.io.Serializable read_value()
	{
		return read_value(read_long(), null, null, null, null);
	}

	/**
	 * Read a value from a CDR stream
	 */
	public java.io.Serializable read_value( String rep_id )
	{
		return read_value(read_long(), rep_id, null, null, null);
	}

	/**
	 * Read a value from a CDR stream
	 */
	public java.io.Serializable read_value( Class clz )
	{
		return read_value(read_long(), null, clz, null, null);
	}

	/**
	 * Read a value from a CDR stream
	 */
	public java.io.Serializable read_value( org.omg.CORBA.portable.BoxedValueHelper boxhelp )
	{
		return read_value(read_long(), null, null, boxhelp, null);
	}

	/**
	 * Implementation of value reading.
	 */
	private java.io.Serializable read_value( int tag, String arg_repo_id, Class clz,
		org.omg.CORBA.portable.BoxedValueHelper boxhelp, java.io.Serializable target )
	{

		// Check if null value
		if ( tag == 0 )
			return null;

		// Check for indirection
		if ( tag == 0xffffffff )
		{
			int loc = _index + read_long();

			Object val = _value_cache.get(new Integer(loc));

			if(val instanceof org.omg.CORBA.MARSHAL)
				cancel((org.omg.CORBA.MARSHAL)val);

			if(val instanceof java.io.Serializable)
				return ( java.io.Serializable )val;

			cancel(new org.omg.CORBA.portable.IndirectionException(loc));
		}

		// Check value validity
		if ( !(( tag >= 0x7fffff00 ) && ( tag <= 0x7fffffff )) )
			cancel(new org.omg.CORBA.MARSHAL("Invalid value tag",
				IIOPMinorCodes.MARSHAL_VALUE, CompletionStatus.COMPLETED_MAYBE));

		// store the offset of the tag for when we have the object
		int tag_index = _index - 4;

		// Check for codebase
		String codeBase = null;
		if ( ( tag & CODEBASE ) == CODEBASE )
		{
			if(peek_long() == 0xffffffff)
			{
        read_long();
				int loc = _index + read_long();

				Object val = _value_cache.get(new Integer(loc));

				if(val == null || !(val instanceof String))
					cancel(new org.omg.CORBA.portable.IndirectionException(loc));

				codeBase = (String)val;
			}
			else
			{
				int codebase_index = _index;
				codeBase = read_string();
				_value_cache.put(new Integer(codebase_index), codeBase);
			}
		}

		String [] ids = null;

		// breakable block
    readids:
		if((tag & MASK_TYPE_INFORMATION) != NO_TYPE_INFORMATION) {
			if((tag & MASK_TYPE_INFORMATION) == MULTIPLE_TYPE_INFORMATION) {
        int count = read_long();

				if(count == 0xFFFFFFFF) {
					int loc = _index + read_long();

					Object val = _value_cache.get(new Integer(loc));

					if(val == null || !(val instanceof String[]))
						cancel(new org.omg.CORBA.portable.IndirectionException(loc));

					ids = (String[])val;
					_value_cache.put(new Integer(_index - 4), ids);
					break readids;
				}
				else
				{
					ids = new String[count];
					_value_cache.put(new Integer(_index - 4), ids);
				}
			}
			else
				ids = new String[1];

			for(int i = 0; i < ids.length; ++i) {
				if(peek_long() == 0xFFFFFFFF)
        {
          read_long();
					int loc = _index + read_long();

					Object val = _value_cache.get(new Integer(loc));

					if(val == null || !(val instanceof String))
						cancel(new org.omg.CORBA.portable.IndirectionException(loc));

					ids[i] = (String)val;
					_value_cache.put(new Integer(_index - 4), ids[i]);
				}
				else
				{
					alignment(4);
					int type_index = _index;
					ids[i] = read_string();
					_value_cache.put(new Integer(type_index), ids[i]);
				}
			}
		}

		boolean old_in_chunked_value = _in_chunked_value;

		int enclosing_level = _value_level;
		if((tag & CHUNK) == CHUNK) {
			int len = peek_long();
      if(!(( len >= 0x7fffff00 ) && ( len <= 0x7fffffff )))
				begin_value_chunk(read_ulong());
			_in_chunked_value = true;
		}
		++_value_level;

		_value_indirect = tag_index;

		// unmarshall
		try {
			if(boxhelp != null) {
				// unmarshal using box helper.
				java.io.Serializable ret = boxhelp.read_value(this);
				_value_cache.put( new Integer( _value_indirect ), ret );
				_value_indirect = -1;
				return ret;
			} else if(target != null) {
				// direct unmarshal case. This is nonstandard.
				return read_value(target);
			} else {
				// default unmarshal case.
				java.io.Serializable ret = null;
				if(ids != null)
				{
					for(int i = 0; i < ids.length; ++i)
					{
						// try value factory first
						if((ret = read_value_withtype(ids[i], codeBase, clz)) != null)
							break;

						// can't find factory for formal paramter type give up.
						if(arg_repo_id != null && ids[i].equals(arg_repo_id))
							break;
					}
				}
				else if(arg_repo_id != null)
					ret = read_value_withtype(arg_repo_id, codeBase, clz);

				if(ret == null) {
					org.omg.CORBA.MARSHAL ex = new org.omg.CORBA.MARSHAL("Unable to locate value factory", org.omg.CORBA.OMGVMCID.value | 1, CompletionStatus.COMPLETED_YES);

					// store marshal exception in indirection table. Missing value
					// factories are ignored for the values within continuation
					// chunks since the data is discarded anyhow, however if somthing
					// indirects to the value the exception gets raised. If this
					// happens within a continuation chunk the exception still gets ignored.
					_value_cache.put(new Integer( _value_indirect ), ex );
					_value_indirect = -1;

					if(_continueLevel > 0)
						throw ex;
					else
						cancel(ex);
				}

				if(clz != null && !clz.isInstance(ret))
				{
					cancel(new org.omg.CORBA.BAD_PARAM("Returned type does not match expected type",
						IIOPMinorCodes.BAD_PARAM_VALUE_CLASS, CompletionStatus.COMPLETED_YES));
				}

				return ret;
			}
		}
		finally {
			if((tag & CHUNK) == CHUNK) {
				// skip over all remaining chunks for this value
				while(_value_level > enclosing_level) {
					if(_encaps_remain > 0)
						force_skip(_encaps_remain);

					_pending_value_reopen = false;

					int endtag = read_long();
					if(endtag >= 0x7fffff00 && endtag <= 0x7FFFFFFF) {
						// value header for contained value. We do have to read
						// these enclosed values as they may be indirected to later
						// on in the buffer.

						int exception_index = _index - 4;
						try {
							++_continueLevel;
							read_value(endtag, null, null, null, null);
						}
						catch(org.omg.CORBA.MARSHAL ex) {
							// ignore can't find value factory marshal exceptions
							// we are ignoring the data anyhow.
							if(ex.minor != (org.omg.CORBA.OMGVMCID.value | 1))
								throw (org.omg.CORBA.MARSHAL)ex.fillInStackTrace();
						}
						finally {
							--_continueLevel;
						}
					}	else if(endtag > 0) {
						// continuation state chunk
						begin_value_chunk(endtag);
					} else {
						// termination chunk
            _value_level = -endtag - 1;
					}
				}
				_pending_value_reopen = _in_chunked_value = old_in_chunked_value;
			} else {
        // non-chunked values terminate silently but may be terminated
				// by contained value chunk.
        _value_level = enclosing_level;
      }
		}
	}

	private void begin_value_chunk(int len) {
		_encaps_stack.addLast(new Integer(_encaps_remain));
		_encaps_stack.addLast(new Integer(_index));
		_encaps_stack.addLast(new Boolean(true));

		_encaps_remain = len;
	}

	private java.io.Serializable read_value_withtype(String repo_id, String codeBase, Class clz) {
		org.omg.CORBA.portable.ValueFactory factory;
		if((factory = ((org.omg.CORBA_2_3.ORB)_orb).lookup_value_factory(repo_id)) != null)
			return factory.read_value(this);

		java.io.Serializable ret;

		// try extended unmarshal
		if((ret = value_extended_unmarshal(_value_indirect, clz, repo_id, codeBase, null )) != null) {
      // add the indirect again, just to make sure.
      _value_cache.put( new Integer( _value_indirect ), ret );
			return ret;
		}

		// try value boxed marshal.
		org.omg.CORBA.portable.BoxedValueHelper boxhelp = null;
		try {
			String boxname = RepoIDHelper.idToClass(repo_id, RepoIDHelper.TYPE_HELPER);
			boxhelp = (org.omg.CORBA.portable.BoxedValueHelper)Class.forName(boxname).newInstance();
		}
		catch(Exception ex) {}
		if(boxhelp != null) {
			ret = boxhelp.read_value(this);
			_value_cache.put( new Integer( _value_indirect ), ret );
			_value_indirect = -1;
			return ret;
		}

		// try loading a remote factory.
		if(codeBase != null && (factory = loadFactoryWithID( repo_id, codeBase )) != null)
			return factory.read_value(this);

		return null;
	}

	/**
	 * Read value state from the stream. Called externally by factories
	 * with uninitialized value. One of the other read_value functions
	 * always appears above it in the call stack.
	 */
	public java.io.Serializable read_value( java.io.Serializable value )
	{
		// nonstandard direct unmashaling to valuetype
		if(_value_indirect < 0)
			return read_value(read_long(), null, null, null, value);

		_value_cache.put( new Integer( _value_indirect ), value );

		// Is a custom value ?
		if ( value instanceof org.omg.CORBA.portable.CustomValue )
			((org.omg.CORBA.portable.CustomValue)value).unmarshal(new org.openorb.CORBA.DataInputStream(this));
		else if ( value instanceof org.omg.CORBA.portable.StreamableValue )
			((org.omg.CORBA.portable.StreamableValue)value)._read(this);
		else
			cancel(new org.omg.CORBA.BAD_PARAM("Unable to read value into class",
				IIOPMinorCodes.BAD_PARAM_VALUE_CLASS, CompletionStatus.COMPLETED_YES));

		return value;
	}

	/**
	 * This function should be overloaded by base types to allow marshaling
	 * of extended value types, RMI over IIOP for example. It should return null
	 * if the repository ID or class is not unmarshaled by the function. All
	 * calls to this function with an offset equal to a previous value must
	 * not unmarshal from the stream.
	 */
	protected java.io.Serializable value_extended_unmarshal(int offset, Class clz, String repo_id, String codeBase, RunTime runTime ) {
		return null;
	}

	/**
	 * Read an abstract interface
	 *
	 * @return the readed object ( An CORBA.Object or a value type )
	 */
	public Object read_abstract_interface()
	{
		if ( read_boolean() )
			return read_Object();
		else
			return read_value();
	}

	/**
	 * Read an abstract interface
	 *
	 * @param clz the stub class for an object
	 * @return the readed object ( An CORBA.Object or a value type )
	 */
	public Object read_abstract_interface( Class clz )
	{
		if ( read_boolean() )
			return read_Object(clz);
		else
			return read_value();
	}

	// --------------------------------------------------------------------------
	// Implementation specific operations
	// --------------------------------------------------------------------------

	/**
	 * internal function. Called after every read_ operation. Pops
	 * encapsulation levels and increments index.
	 */
	private void postread(int len)
	{
		_index += len;

		if(_encaps_remain > 0) {
			_encaps_remain -= len;

			while(_encaps_remain == 0) {
				boolean value_chunk = ((Boolean)_encaps_stack.removeLast()).booleanValue();

				int enclen = _index - ((Integer)_encaps_stack.removeLast()).intValue();
				_encaps_remain = ((Integer)_encaps_stack.removeLast()).intValue() - enclen;

				if(value_chunk)
				{
					// value chunk ends.
					_pending_value_reopen = true;
				}
				else
				{
					// encapsulation ends
					_swap = ((Boolean)_encaps_stack.removeLast()).booleanValue();
					if(!_encaps_stack.isEmpty())
						_in_chunked_value = ((Boolean)_encaps_stack.getLast()).booleanValue();
					_pending_encaps_close++;
				}
				// check for top layer encapsulation
				if(_encaps_remain < 0)
					_encaps_remain = -1;
			}
		}
	}

	/**
	 * Load a OBV factory from its ID.
	 */
	protected org.omg.CORBA.portable.ValueFactory loadFactoryWithID( String id, String url )
	{
		String factoryName = RepoIDHelper.idToClass(id, RepoIDHelper.TYPE_DEFAULT_FACTORY);

		Class clzHelper;

		try {
      clzHelper = loadClassFromURL( factoryName, url );

			if ( url != null )
			  clzHelper = loadClassFromURL( factoryName, url );
			else
			  clzHelper = Class.forName(factoryName);

			if ( clzHelper != null )
				return ( org.omg.CORBA.portable.ValueFactory ) clzHelper.newInstance();
		}
		catch(Exception ex) {
		}
		return null;
	}

	/**
	 * Try to load a class from an URL site.
	 */
	private Class loadClassFromURL( String className, String url )
	{
    try {
      if(url == null)
        return Class.forName(className);
      else
      {
        ClassLoader loader;

        loader = (ClassLoader)_value_cache.get(url);

        if(loader == null) {
          java.util.StringTokenizer token = new java.util.StringTokenizer( url, " " );

          java.net.URL [] urls = new java.net.URL[token.countTokens()];

          int pos = 0;
          for(int i = 0; i < urls.length; ++i)
          {
            try
            {
              urls[pos] = new java.net.URL(token.nextToken());
              ++pos;
            }
            catch(java.net.MalformedURLException ex) {}
          }

          if(pos != urls.length) {
						java.net.URL [] old = urls;
            urls = new java.net.URL[pos];
            System.arraycopy(old, 0, urls, 0, pos);
          }

          loader = new java.net.URLClassLoader(urls);

          _value_cache.put(url, loader);
        }

        Class clz = loader.loadClass( className );

        return clz;
      }
    }
    catch ( ClassNotFoundException ex )
    {
      org.openorb.util.Trace.high(this, "Unable to find class from an URL : "  + className );
    }

    return null;
  }
}

