/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: IIOPClientChannel.java,v 1.14 2001/08/13 03:47:33 wood Exp $
 *
 * Date         Author  Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/
package org.openorb.iiop;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Iterator;

import java.net.*;
import java.io.*;

import org.openorb.net.*;
import org.openorb.io.*;
import org.omg.BiDirPolicy.*;

/**
 * Implements the {@link org.openorb.net.ClientChannel} interface for IIOP.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.14 $ $Date: 2001/08/13 03:47:33 $
 */
public class IIOPClientChannel implements ClientChannel {

    /**
     * Construct an initiating client channel.
     * @param protocol The client protocol.
     * @param transp Message transport.
     * @param serverProtocol non-null is a prerequisite for bidirectional use.
     * @param tcsc char codeset.
     * @param tcsw wchar codeset.
     */
    IIOPClientChannel(IIOPClientProtocol protocol, Transport transp,
		      IIOPServerProtocol serverProtocol, int tcsc, int tcsw) {
	_protocol = protocol;
	_orb = _protocol.orb();
	_clientManager = _protocol.getClientManager();
	_socketQueue = new SocketQueue(transp);
	_socketQueue.setClientChannel(this);

	// create sync variables
	_syncState = new Object();

	_serverProtocol = serverProtocol;

	_tcsc = tcsc;
	_tcsw = tcsw;

	try {
	    Class [] cargs = new Class[] { org.omg.CORBA.ORB.class, org.omg.GIOP.Version.class, MarshalBuffer.class };
	    _os_ctor = ((org.openorb.CORBA.ORB)_orb).getLoader().classConstructor("iiop.CDROutputStreamClass", "org.openorb.iiop.CDROutputStream", cargs);
	}
	catch(Exception ex) {
	    throw new org.omg.CORBA.INITIALIZE("Unable to initialize output stream constructor");
	}
	org.openorb.util.Trace.high(this, "created");
    }

    /**
     * Construct a delegate client channel.
     */
    IIOPClientChannel(IIOPClientProtocol protocol, IIOPServerChannel peer) {
	_protocol = protocol;
	_orb = _protocol.orb();
	_clientManager = _protocol.getClientManager();

	_peer = peer;
	_socketQueue = _peer._socketQueue;
	_socketQueue.setClientChannel(this);
	_syncState = _peer._syncState;

  	_state = STATE_CONNECTED;
	_isDelegated = true;

	_tcsc = _peer._tcsc;
	_tcsw = _peer._tcsw;

	try {
	    Class [] cargs = new Class[] { org.omg.CORBA.ORB.class, org.omg.GIOP.Version.class, MarshalBuffer.class };
	    _os_ctor = ((org.openorb.CORBA.ORB)_orb).getLoader().classConstructor("iiop.CDROutputStreamClass", "org.openorb.iiop.CDROutputStream", cargs);
	}
	catch(Exception ex) {
	    throw new org.omg.CORBA.INITIALIZE("Unable to initialize output stream constructor");
	}
    }

    // all the variables here which are package protected are so to allow
    // access / sharing by IIOPServerChannel when they are peered.

    private static final byte [] RESERVED = { 0, 0, 0 };

    private org.omg.CORBA.ORB _orb;
    private IIOPClientProtocol _protocol;
    private ClientManager _clientManager;

    // synchronization for all the vars below to _syncRecv
    // shared with peer if there is one.
    Object _syncState;

    /* all the below are synchronized on state sync */
    int _state = STATE_PAUSED;

    // substates.
    private boolean _firstConnect = true;
    private boolean _inPause = false;
    boolean _pendClose = false;
    private boolean _pendSC = false;

    SocketQueue _socketQueue;
    private org.omg.IIOP.Version _version = new org.omg.IIOP.Version((byte)1, (byte)0);

    org.omg.CORBA.SystemException _closeException = null;

    private short _disposition = org.omg.GIOP.KeyAddr.value;

    Map _activeRequests = new HashMap();
    int _channelAge = 0;

    int _tcsc;
    int _tcsw;

    private static final int CONN_SC_TOTAL = 2;

    // the service contexts
    private static final int CONN_SC_CODESETS = 0;
    private static final int CONN_SC_BIDIR = 1;

    // the sending status.
    private static final int CONN_SC_STATUS_UNSENT = 0;
    private static final int CONN_SC_STATUS_PEND = 1;
    private static final int CONN_SC_STATUS_SENT = 2;

    private int [] _scSent = null;

    // stuff for bidir
    private boolean _isDelegated = false;
    private IIOPServerProtocol _serverProtocol;

    IIOPServerChannel _peer;

    private java.lang.reflect.Constructor _os_ctor;

    public String toString() {
	return "ClientChannel: " + _socketQueue.toString();
    }

    /**
     * Return the GIOP/IIOP version
     */
    public org.omg.IIOP.Version version() {
	return _version;
    }

    /**
     * The selected char codeset.
     */
    public int getTCSC() {
	return _tcsc;
    }

    /**
     * The selected wide codeset.
     */
    public int getTCSW() {
	return _tcsw;
    }

    // connection state.

    public org.omg.CORBA.ORB orb() {
	return _orb;
    }

    /**
     * Connection state. Note that state values are 0x1000000 apart,
     * Allowing this value to be ORed with the binding priority.
     */
    public int state() {
	if(_isDelegated) {
	    synchronized(_syncState) {
		if(_peer != null)
		    return STATE_CONNECTED;
		return STATE_PAUSED;
	    }
	}
	return _state;
    }

    boolean isDelegated() {
	return _isDelegated;
    }

    /**
     * Active request count. This is the number of requests which have not
     * yet been sent or are still expecting a reply.
     */
    public int active_requests() {
	synchronized(_syncState) {
	    return _activeRequests.size() + ((_peer != null)?_peer._activeRequests.size():0);
	}
    }

    /**
     * Indication of channel age. This will result in a call to peek_request_id
     * if active_requests would be non-zero, otherwise it returns whatever
     * peek_request_id returned the last time active_requests dropped to zero.
     */
    public int channel_age() {
	synchronized(_syncState) {
	    if(!_activeRequests.isEmpty() || (_peer != null && !_peer._activeRequests.isEmpty()))
		return RequestIDAllocator.peek_request_id();
	    if(_peer != null && _channelAge < _peer._channelAge)
		return _peer._channelAge;
	    return _channelAge;
	}
    }

    /**
     * Open channel for request. _syncState should be owned.
     */
    private void request_open(org.omg.CORBA.Object target, Address address)
    {
	switch(_state) {
	case STATE_CONNECTED:
	    return;
	case STATE_CLOSED:
	    throw _closeException;
	}

	// check the rebind mode.
	if(false && !_firstConnect) {
	    // TODO: check the rebind mode
	    throw new org.omg.CORBA.REBIND(0, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
	}

	// check to ensure peer has not closed between creating request
	// and sending it.
	if(_isDelegated && _peer == null)
	    throw new org.omg.CORBA.TRANSIENT(0, org.omg.CORBA.CompletionStatus.COMPLETED_NO);

	// may throw exception.
	_socketQueue.open();

	_state = STATE_CONNECTED;
	_firstConnect = false;
	_pendClose = false;
	_closeException = null;
	connectionSCReset();
	_peer = null;

	_clientManager.register_channel(this);

	org.openorb.util.Trace.high(this, "opened");
	return;
    }

    /**
     * Begin the pausing sequence, enter the PAUSED state. As soon as
     * work_pending returns false this will close the connection.
     */
    public void pause() {
	if(_isDelegated) {
	    _peer.soft_close(false);
	    return;
	}

	synchronized(_syncState) {
	    // ignore pause request if there are active requests.
	    if(active_requests() > 0)
		return;

	    // can't pause if we are not in a pausable state.
	    if(_state != STATE_CONNECTED)
		return;

	    org.openorb.util.Trace.high(this, "paused");

	    _state = STATE_PAUSED;

	    // close the peer.
	    if(_peer != null) {
		_peer._state = IIOPServerChannel.STATE_CLOSED;
		_peer._peer = null;
		_peer = null;
	    }

	    _inPause = true;
	}

	// stop the worker threads.
	_clientManager.unregister_channel(this);

	_socketQueue.close();

	synchronized(_syncState) {
	    _inPause = false;
	    _syncState.notifyAll();
	}
    }

    /**
     * Respond to a server CLOSE_CONNECTION message. This will cancel all active
     * requests but allow them to be resent on a new connection. This will also
     * be called by a server with this as a delegate.
     */
    void server_pause(org.omg.CORBA.SystemException ex) {
	IIOPServerChannel peer = null;
	ServerRequest [] serverRequests = null;
	synchronized(_syncState) {
	    // can't pause if we are not in a pausable state.
	    if(_state != STATE_CONNECTED) {
		org.openorb.util.Trace.high(this, "already closed");
		return;
	    }

	    _state = STATE_PAUSED;

	    // discard the peer. There should be no active requests for the peer
	    // (or the server would not be sending close connection messages)
	    if(_peer != null) {
		if(!_peer._activeRequests.isEmpty()) {
		    // TODO: this shouldn't happen, but there's noone to
		    // report the error to, the client side has probably
		    // shut down by now.
		    serverRequests = new ServerRequest[_peer._activeRequests.size()];
		    _peer._activeRequests.values().toArray(serverRequests);
		    _peer._activeRequests.clear();
		}

		peer = _peer;

		_peer._state = IIOPServerChannel.STATE_CLOSED;
		_peer._peer = null;
		_peer = null;
	    }

	    _inPause = true;
	}

	Object name;

	// stop the worker threads.
	if(_isDelegated) {
	    name = peer;
	    peer._serverManager.unregister_channel(peer);
	} else {
	    name = this;
	    _clientManager.unregister_channel(this);
	}

	_socketQueue.close();

	IIOPClientRequest [] clientRequests = null;
	synchronized(_syncState) {
	    _inPause = false;
	    _syncState.notifyAll();

	    // cancel any active requests with a transient exception.
	    // the cancel gets called outside of the sync block.
	    if(!_activeRequests.isEmpty()) {
		clientRequests = new IIOPClientRequest[_activeRequests.size()];
		_activeRequests.values().toArray(clientRequests);
		_activeRequests.clear();
	    }

	    org.openorb.util.Trace.high(name, _isDelegated?"paused by client":"paused by server");
	}
	if(clientRequests != null)
	    for(int i = 0; i < clientRequests.length; ++i)
		clientRequests[i].cancel(ex);
	if(serverRequests != null)
	    for(int i = 0; i < serverRequests.length; ++i)
		serverRequests[i].client_cancel();
    }

    /**
     * Change to the CLOSED state and reject new requests
     * by throwing a system exception with status COMPLETED_NO.
     * If kill_requests is true, call cancel on any active
     * ClientRequests, otherwise wait until they complete before closing.
     */
    public void close(boolean kill_requests, org.omg.CORBA.SystemException ex) {
	if(_isDelegated) {
	    if(kill_requests)
		_peer.close();
	    else
		_peer.soft_close(_pendClose);
	    return;
	}

	IIOPClientRequest [] clientRequests = null;
	ServerRequest [] serverRequests = null;
	synchronized(_syncState) {
	    // wait for any pauses to complete
	    if(_state == STATE_PAUSED) {
		boolean interrupted = false;
		while(_inPause) {
		    try {
			_syncState.wait();
		    }
		    catch(InterruptedException ir) {
			interrupted = true;
		    }
		}
		if(interrupted)
		    Thread.currentThread().interrupt();
		_closeException = ex;
		_state = STATE_CLOSED;
		return;
	    } else if(_state == STATE_CLOSED) {
		return;
	    }

	    _state = STATE_CLOSED;
	    _closeException = ex;

	    // check for clean shutdown.
	    if(!kill_requests && active_requests() > 0) {
		_pendClose = true;
		if(_peer != null)
		    _peer._pendClose = true;
		return;
	    }

	    clientRequests = new IIOPClientRequest[_activeRequests.size()];
	    _activeRequests.values().toArray(clientRequests);
	    _activeRequests.clear();

	    if(_peer != null) {
		serverRequests = new ServerRequest[_peer._activeRequests.size()];
		_peer._activeRequests.values().toArray(serverRequests);
		_peer._activeRequests.clear();

		_peer._state = IIOPServerChannel.STATE_CLOSED;
		_peer._peer = null;
		_peer = null;
	    }
	}

	// stop the worker threads.
	_clientManager.unregister_channel(this);

	_socketQueue.close();

	org.openorb.util.Trace.high(this, "closed");

	// cancel all the active requests.
	if(clientRequests != null)
	    for(int i = 0; i < clientRequests.length; ++i)
		clientRequests[i].cancel(ex);
	if(serverRequests != null)
	    for(int i = 0; i < serverRequests.length; ++i)
		serverRequests[i].client_cancel();
    }

    /**
     * Create a request. If this is the first request on this channel
     * then client_connect will be called on all ChannelInterceptor
     * before returning the request. This may throw a system exception if the
     * channel cannot establish a connection for some reason, for example
   * INV_POLICY if client side policies prevent a successfull invocation,
     * COMM_FAILURE if a communication problem occours, or REBIND if channel
     * is temporaraly closed and a NO_REBIND policy is in effect.
   *
   * @param target The target of the request.
   * @param address The target address. If the target has been redirected
   *            this may not correspond to the target's ior.
   */
    public ClientRequest create_request(org.omg.CORBA.Object target,
					Address address, Object tpAssc,
					String operation,
					boolean response_expected)
	throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	//System.out.println("Before create");
	IIOPClientRequest request = new IIOPClientRequest(request_id, target, address, tpAssc, this, operation, response_expected);
	//System.out.println("BEfore synch");
	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }


    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }


    /* Added by Diana on 2 feb. 2002*/

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.IOP.ServiceContext one_sc,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,one_sc,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }


    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,org.omg.IOP.ServiceContext one_sc,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,request_start_time,one_sc,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }
    /* Modif. on Oct. 28 by Diana*/

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,request_start_time,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }


    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.Dynamic.Parameter[] arguments,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,arguments,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }
    /* new version by Diana Oct. 10 2001*/

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions, org.omg.CORBA.Any result,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,arguments,exceptions,result,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }


    /* new version by Diana Oct. 28 2001*/

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions, org.omg.CORBA.Any result,Address address, Object tpAssc,String operation,boolean response_expected) throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, retention_id,client_id,request_start_time,arguments,exceptions,result,target, address, tpAssc, this, operation, response_expected);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "request #" + request_id + " created");

	return request;
    }

    /**
   * Create a locate request. This may throw a system exception if the
   * channel cannot establish a connection for some reason, for example
   * INV_POLICY if client side policies prevent a successfull invocation,
   * COMM_FAILURE if a communication problem occours, or REBIND if channel
   * is temporaraly closed and a NO_REBIND policy is in effect.
   *
   * @param target The target of the request.
   * @param address The target address. If the target has been redirected
   *            this may not correspond to the target's ior.
   */
    public ClientRequest create_locate_request(org.omg.CORBA.Object target,
					       Address address, Object tpAssc)
	throws RebindChannelException
    {
	int request_id = RequestIDAllocator.get_request_id() << 1;
	if(_isDelegated)
	    request_id--;

	IIOPClientRequest request = new IIOPClientRequest(request_id, target, address, tpAssc, this);

	synchronized(_syncState) {
	    check_rebind(address);

	    _activeRequests.put(new Integer(request_id), request);
	}

	org.openorb.util.Trace.medium(this, "locate request #" + request_id + " created");

	return request;
    }

    /**
     * Check for channel rebind. _syncState should be owned.
   */
    private void check_rebind(Address address)
	throws RebindChannelException
    {
	if(_isDelegated && _peer == null) {
	    _protocol.rebindBidirDelegate(this, (IIOPAddress)address);
	    org.openorb.util.Trace.assertTrue(false);
	    return;
	}
    }

    /**
     * Send a cancel message for the given request or just discard it.
   */
    boolean cancel_request(IIOPClientRequest request, boolean sendMesg) {
	int req_id = request.request_id();
	boolean closeChan = false;
	synchronized(_syncState) {
	    if(_state == STATE_CLOSED)
		return false;

	    if(_pendSC && !request.is_locate()) {
		connectionSCCanceled();
		_pendSC = false;
		_syncState.notifyAll();
	    }

	    if(_activeRequests.isEmpty() || _activeRequests.remove(new Integer(req_id)) == null)
		return false;
	    if(_activeRequests.isEmpty()) {
		_channelAge = RequestIDAllocator.peek_request_id();
		closeChan = _pendClose && ( _peer == null || _peer._activeRequests.isEmpty());
	    }
	}

	org.openorb.util.Trace.medium(this, "request #" + req_id + " canceled");

	// if a close is pending just close the channel, no need to send a cancel
	// message.
	if(closeChan) {
	    if(_isDelegated)
		_peer.soft_close(false);
	    else
		close(false, _closeException);
	    return true;
	}

	// send cancel message
	if(sendMesg) {
	    byte [] cancel = new byte[] {
		(byte)'G', (byte)'I', (byte)'O', (byte)'P',
		    _version.major, _version.minor,
		    org.omg.GIOP.MsgType_1_1._CancelRequest, 4,
		    (byte) (req_id >>> 24), (byte) (req_id >>> 16),
		    (byte) (req_id >>> 8), (byte) req_id
		    };
	    StorageBuffer buf = new StorageBuffer(cancel, 0, 12);

	    _socketQueue.send(buf, request.get_transport_association());
	}

	return true;
    }

    /**
     * Create output stream for request or throw a system exception
     */

    /* if want to modify something about the arguments sent, so that they are known to the
       server, at the interceptor level, do a modification in method begin_marshal*/

    org.omg.CORBA_2_3.portable.OutputStream begin_marshal(IIOPClientRequest req) {
	org.omg.IIOP.Version iiopvers = ((IIOPAddress)req.address()).get_version();
	org.omg.GIOP.Version req_vers = new org.omg.GIOP.Version((byte)1, (byte)0);

	short req_disposition;

	boolean addCodesets = false;
	synchronized(_syncState) {
	    // the channel's version will match the highest iiop version of
	    // all requests issued through it.
	    if(_version.minor < iiopvers.minor)
		_version.minor = iiopvers.minor;
	    req_vers.minor = _version.minor;

	    boolean isRequest = !req.is_locate();

	    if(isRequest)
		while(_inPause || _pendSC) {
		    try {
			_syncState.wait();
		    }
		    catch(InterruptedException ex) {
			// TODO: containing request has been canceled, cancel this request.
		    }
		}

	    request_open(req.target(), req.address());

	    if(isRequest)
		_pendSC = connectionSCAdd(req);

	    req_disposition = _disposition;
	}

	MarshalBuffer buf = new MarshalBuffer(new MarshalListener(), req);
	CDROutputStream os;
	try {
	    Object [] args = new Object[3];
	    args[0] = req.orb();
	    args[1] = req_vers;
	    args[2] = buf;
	    os = (CDROutputStream)_os_ctor.newInstance(args);
	}
	catch(Exception ex) {
	    throw new org.omg.CORBA.INITIALIZE("Unable to create CDROutputStream class");
	}
	os.setCodesets(_tcsc, _tcsw);

	int req_id = req.request_id();
	if(req.is_request()) {
	    new HeaderBlock((byte)org.omg.GIOP.MsgType_1_1._Request, req_id, os);

	    switch(req_vers.minor) {
	    case 0:
	    case 1:
		org.omg.IOP.ServiceContextListHelper.write(os, req.get_request_service_contexts());
		os.write_long(req_id);
		os.write_boolean(req.response_expected());
		// don't need to explicitly write the reserved bytes,
		// object key's length is 4-aligned anyhow.
		org.omg.CORBA.OctetSeqHelper.write(os, req.address().getTargetAddress(org.omg.GIOP.KeyAddr.value).object_key());
		os.write_string(req.operation());
		// don't write any principal
		os.write_long(0);
		break;
	    case 2:
		os.write_long(req_id);
		os.allowFragment();
		switch(req.sync_scope()) {
		case 0: /* org.omg.Messaging.SYNC_NONE.value */
		case 1: /* org.omg.Messaging.SYNC_WITH_TRANSPORT.value */
		    os.write_octet((byte)0);
		    break;
		case 2: /* org.omg.Messaging.SYNC_WITH_SERVER.value */
		    os.write_octet((byte)1);
		    break;
		case 3: /* org.omg.Messaging.SYNC_WITH_TARGET.value */
		    os.write_octet((byte)3);
		    break;
		}
		// reserved. This does need to be written, the discriminator of target
		// address is a short
		os.write_octet_array(RESERVED, 0, 3);
		org.omg.GIOP.TargetAddressHelper.write(os, req.address().getTargetAddress(req_disposition));
		os.write_string(req.operation());
		org.omg.IOP.ServiceContextListHelper.write(os, req.get_request_service_contexts());
		os.pending_alignment(8);
		break;
	    }
	} else {
	    new HeaderBlock((byte)org.omg.GIOP.MsgType_1_1._LocateRequest, req_id, os);

	    os.write_long(req_id);
	    switch(req_vers.minor) {
	    case 0:
	    case 1: {
		byte [] oid = req.address().getTargetAddress(org.omg.GIOP.KeyAddr.value).object_key();
		org.omg.CORBA.OctetSeqHelper.write(os, oid);
	    } break;
	    case 2:
		os.allowFragment();
		org.omg.GIOP.TargetAddressHelper.write(os, req.address().getTargetAddress(req_disposition));
		break;
	    }
	}

	return os;
    }

    /**
     * Add any neccicary conneciton SCs, and set the status to pending.
     * @returns true if a connection SC is now pending
     */
    protected boolean connectionSCAdd(IIOPClientRequest req) {
	if(_scSent == null)
	    return false;

	boolean addSent = false;
	boolean addPend = false;

	// codesets SC
	if(_scSent[CONN_SC_CODESETS] == CONN_SC_STATUS_UNSENT && _version.minor > 0) {
	    if(_tcsc != 0 || _tcsw != 0) {
		byte [] data = new byte[] {
		    0,0,0,0,
			(byte)(_tcsc >>> 24), (byte)(_tcsc >>> 16),
			(byte)(_tcsc >>> 8 ), (byte)_tcsc,
			(byte)(_tcsw >>> 24), (byte)(_tcsw >>> 16),
			(byte)(_tcsw >>> 8 ), (byte) _tcsw
			};
		org.omg.IOP.ServiceContext _codesetSC
		    = new org.omg.IOP.ServiceContext(org.omg.IOP.TAG_CODE_SETS.value, data);
		req.add_request_service_context(_codesetSC, true);
		_scSent[CONN_SC_CODESETS] = CONN_SC_STATUS_PEND;
		addPend = true;
	    } else {
		_scSent[CONN_SC_CODESETS] = CONN_SC_STATUS_SENT;
		addSent = true;
	    }
	}

	// bidir SC
	if(_scSent[CONN_SC_BIDIR] == CONN_SC_STATUS_UNSENT) {
	    org.omg.IOP.ServiceContext sc = _protocol.getBiDirSC();
	    if(sc != null) {
		BidirectionalPolicy pol = null;
		try {
		    pol = BidirectionalPolicyHelper.narrow(req.get_request_policy(BIDIRECTIONAL_POLICY_TYPE.value));
		}
		catch(org.omg.CORBA.INV_POLICY ex) {
		}

		if(pol != null && pol.value() == BOTH.value) {
		    _peer = new IIOPServerChannel(_serverProtocol.getServerManager(), this, _protocol._codec);
		    org.openorb.util.Trace.high(this, "became bidirectional with peer " + _peer.toString());


		    req.add_request_service_context(sc, true);

		    _scSent[CONN_SC_BIDIR] = CONN_SC_STATUS_PEND;
		    addPend = true;
		}
	    } else {
		_scSent[CONN_SC_BIDIR] = CONN_SC_STATUS_SENT;
		addSent = true;
	    }
	}

	if(addPend)
	    return true;

	if(addSent) {
	    boolean allSent = true;
	    for(int i = 0; i < _scSent.length; ++i)
		if(!(allSent = _scSent[i] == CONN_SC_STATUS_SENT))
		    break;
	    if(allSent)
		_scSent = null;
	}

	return false;
    }

    /**
     * Change the status of any pending connection SCs to sent.
   */
    protected void connectionSCSent() {
	if(_scSent != null) {
	    boolean allSent = true;
	    for(int i = 0; i < _scSent.length; ++i) {
		if(_scSent[i] != CONN_SC_STATUS_UNSENT)
		    _scSent[i] = CONN_SC_STATUS_SENT;
		else
		    allSent = false;
	    }
	    if(allSent)
		_scSent = null;
	}
    }

    /**
     * Change the status of any pending connection SCs to unsent.
     * @returns true if connection SC sent status changed.
     */
    protected void connectionSCCanceled() {
	if(_scSent != null)
	    for(int i = 0; i < _scSent.length; ++i)
		if(_scSent[i] == CONN_SC_STATUS_PEND)
		    _scSent[i] = CONN_SC_STATUS_UNSENT;
    }

    /**
     * Change the status of all connection SCs to unsent.
   */
    protected void connectionSCReset() {
	_scSent = new int[CONN_SC_TOTAL];
    }

    private class MarshalListener implements MarshalBuffer.Listener {
	private boolean _fragSent = false;

	public void availIncreaced(MarshalBuffer buffer, int available, Object cookie) {
	    while(available > SocketQueue.MAX_FRAG_SIZE) {
		_fragSent = true;

		IIOPClientRequest request = (IIOPClientRequest)cookie;
		if(request.state() != request.STATE_MARSHAL)
		    return;
		// TODO: set the minor code to somthing
		StorageBuffer buf = buffer.fragment(SocketQueue.MAX_FRAG_SIZE);

		if(_socketQueue.send(buf, request.get_transport_association()))
		    org.openorb.util.Trace.medium(IIOPClientChannel.this, "request #" + request.request_id() + " fragment sent");

		available -= SocketQueue.MAX_FRAG_SIZE;
	    }
	}

	public void bufferClosed(MarshalBuffer buffer, int available, Object cookie) {
	    IIOPClientRequest request = (IIOPClientRequest)cookie;
	    if(request.state() != request.STATE_MARSHAL)
		return;

	    StorageBuffer buf = buffer.lastFragment();
	    if(_socketQueue.send(buf, request.get_transport_association()))
		org.openorb.util.Trace.medium(IIOPClientChannel.this, "request #" + request.request_id() + " last fragment sent");

	    if(_pendSC && !request.is_locate()) {
		synchronized(_syncState) {
		    connectionSCSent();
		    _pendSC = false;
		    _syncState.notifyAll();
		}
	    }
	}

	public void bufferCanceled(MarshalBuffer buffer, org.omg.CORBA.SystemException ex, Object cookie) {
	    IIOPClientRequest request = (IIOPClientRequest)cookie;

	    // must properly cancel the request with a cancel message
	    cancel_request(request, _fragSent);

	    if(request.state() != request.STATE_MARSHAL)
		return;

	    // call cancel on the request. If the request caused this call then it
	    // will be in the complete state and no action will be taken,
	    // otherwise the cancel that gets propogated back through the stream will
	    // not get back to here, it gets filtered out in the cancel op in the
	    // marhshal buffer.
	    ex.completed = org.omg.CORBA.CompletionStatus.COMPLETED_NO;
	    request.cancel(ex);
	    throw ex;
	}
    };

    /**
     * Wait the specified amount of time for an incoming message.
     * @returns false if the channel is closed.
     */
    public boolean recv(int timeout) {
	if(_isDelegated)
	    return false;
	return _socketQueue.receive(timeout);
    }

    /**
     * Donate a thread for recieving messages. This function returns when interrupt
     * is called on the thread or the channel is closed.
   */
    public void run_recv()
    {
	if(_isDelegated)
	    return;
	while(!Thread.interrupted() && _socketQueue.receive(1000));
    }

    int process_reply(byte minor, CDRInputStream is, byte msg_type, boolean fragFollows, BufferSource source, Object tpAssc)
    {
	int reply_status = -1;
	org.omg.IOP.ServiceContext [] reply_service_contexts = null;

	is.setCodesets(_tcsc, _tcsw);

	int req_id = 0;
	IIOPClientRequest request = null;
	boolean closeChan = false;
	try {
	    // read SCs from < IIOP 1.2
	    if(minor < 2 && msg_type == org.omg.GIOP.MsgType_1_1._Reply)
		reply_service_contexts = org.omg.IOP.ServiceContextListHelper.read(is);

	    // find the request
	    req_id = is.read_long();
	    synchronized(_syncState) {
		if(fragFollows)
		    request = (IIOPClientRequest)_activeRequests.get(new Integer(req_id));
		else {
		    request = (IIOPClientRequest)_activeRequests.remove(new Integer(req_id));
		    if(_activeRequests.isEmpty()) {
			_channelAge = RequestIDAllocator.peek_request_id();
			closeChan = _pendClose && ( _peer == null || _peer._activeRequests.isEmpty());
		    }
		}
	    }

	    // missing request for reply? request most likley has been canceled
	    // discard the reply.
	    if(request == null) {
		org.openorb.util.Trace.medium(this,
					      "request #" + req_id  + (fragFollows?"":" last") + " reply fragment discarded, no corresponding request.");
	    } else if(!request.checkReplyTransportAssoc(tpAssc)) {
		// discard anauthenticated fragment.
		request = null;
		org.openorb.util.Trace.medium(this,
					      "request #" + req_id  + (fragFollows?"":" last") + " reply fragment discarded, sender not authorized.");
	    } else {
		// normal request
		org.openorb.util.Trace.medium(this,
					      "request #" + req_id  + (fragFollows?"":" last") + " reply fragment received");

		request.setReplySource(source);
		reply_status = is.read_long();
		//System.out.println("REply status is "+reply_status+" "+req_id);
		if(minor >= 2 && msg_type == org.omg.GIOP.MsgType_1_1._Reply) {
		    //  System.out.println("Read service contexts "+req_id);
		    reply_service_contexts = org.omg.IOP.ServiceContextListHelper.read(is);
		    try {
			is.alignment(8);
		    }
		    catch(org.omg.CORBA.MARSHAL ex1) {
			// thrown with an overread, ignore.
		    }
		}
	    }
	}
	catch(org.omg.CORBA.SystemException ex) {
	    // this will occour if the connection is lost between recieving
	    // fragments of the service contexts or a marshal exception occours
	    // while unmarshaling the service contexts either way no
	    // handle_reply will be called.

	    // If the marshal exception occours while unmarshaling a < 1.2 service
	    // context then there is no easy way of identifying which request's
	    // reply was lost (we could flag all active requests, and kill the last
	    // one which doesn't get a reply, but it's not worth it
	    if(request == null)
		close(true, _closeException);
	    else
		request.cancel(ex);
	    return req_id;
	}
	finally {
	    source.removeWaitingForBufferListener(null);
	}

	// add the reply to the request.
	if(request != null)
	    request.handle_reply(reply_status, reply_service_contexts, is);

	// if a close is pending we can complete it now
	if(closeChan) {
	    if(_isDelegated)
		_peer.soft_close(false);
	    else
		close(false, _closeException);
	}

	return req_id;
    }

    void process_fragment(int req_id, StorageBuffer buf, boolean fragFollows, Object tpAssc) {
	// we have a reply fragment
	IIOPClientRequest request;
	boolean closeChan = false;
	synchronized(_syncState) {
	    if(fragFollows)
		request = (IIOPClientRequest)_activeRequests.get(new Integer(req_id));
	    else {
		request = (IIOPClientRequest)_activeRequests.remove(new Integer(req_id));
		if(_activeRequests.isEmpty()) {
		    _channelAge = RequestIDAllocator.peek_request_id();
		    closeChan = _pendClose && ( _peer == null || _peer._activeRequests.isEmpty());
		}
	    }
	}

	// do the traces
	if(request == null) {
	    org.openorb.util.Trace.medium(this,
					  "request #" + req_id + (fragFollows?"":" last") + " reply fragment discarded, no corresponding request.");
	} else if(!request.checkReplyTransportAssoc(tpAssc)) {
	    org.openorb.util.Trace.medium(this,
					  "request #" + req_id + (fragFollows?"":" last") + " reply fragment discarded, sender not authorized.");
	} else {
	    org.openorb.util.Trace.medium(this,
					  "request #" + req_id + (fragFollows?"":" last") + " reply fragment.");
	    request.getReplySource().addLast(buf, !fragFollows);
	}

	// if a close is pending we can complete it now
	if(closeChan) {
	    if(_isDelegated)
		_peer.soft_close(false);
	    else
		close(false, _closeException);
	}
    }

    /**
     * Finalize method closes channel.
	 */
    protected void finalize()
	throws Throwable
    {
	close(false, new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 4, org.omg.CORBA.CompletionStatus.COMPLETED_NO));
    }

}
