/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: IIOPClientProtocol.java,v 1.7 2001/08/13 03:47:33 wood Exp $
 *
 * Date         Author  Changes
 */

/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/
package org.openorb.iiop;

import java.util.Map;
import org.openorb.util.WeakValueHashMap;

import java.util.Iterator;
import java.util.LinkedList;

import java.net.InetAddress;
import java.net.UnknownHostException;

import org.openorb.net.*;

import org.openorb.adapter.ObjectAdapter;
import org.omg.CONV_FRAME.CodeSetComponentInfo;
import org.omg.CONV_FRAME.CodeSetComponent;
import org.omg.CORBA.CompletionStatus;
import org.omg.IIOP.ListenPoint;

/**
 * Implements the {@link org.openorb.net.ClientProtocol} interface for IIOP.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.7 $ $Date: 2001/08/13 03:47:33 $
 */
public class IIOPClientProtocol implements ClientProtocol {

	/** Creates new IIOPClientProtocol */
	public IIOPClientProtocol(ClientManager clientManager, CDRCodecFactory codec_factory,
	                          CodeSetComponentInfo codesetInfo, TransportClientInitializer tpInit,
                            ListenPoint [] biDirListenPoints) {
		try {
			_codec = codec_factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		}
		catch(org.omg.IOP.CodecFactoryPackage.UnknownEncoding ex) {
			org.openorb.util.Trace.assertTrue(false, "Could not find codec");
		}
		_clientManager = clientManager;
		_orb = _clientManager.orb();

		_codesetInfo = codesetInfo;

		_tpInit = tpInit;

    if(biDirListenPoints != null) {
      try {
        org.omg.IIOP.BiDirIIOPServiceContext scd = new org.omg.IIOP.BiDirIIOPServiceContext(biDirListenPoints);

        org.omg.CORBA.Any any = _orb.create_any();
        org.omg.IIOP.BiDirIIOPServiceContextHelper.insert(any, scd);
        byte [] data = _codec.encode_value(any);

        _biDirSC = new org.omg.IOP.ServiceContext(org.omg.IOP.BI_DIR_IIOP.value, data);
      }
      catch(org.omg.CORBA.UserException ex) {
        throw new org.omg.CORBA.INITIALIZE(ex.toString());
      }
    }
	}

	void setServerProtocol(IIOPServerProtocol protocol) {
		_serverProtocol = protocol;
		_serverManager = protocol.getServerManager();
	}

	private ClientManager _clientManager;

	private org.omg.CORBA.ORB _orb;
	org.omg.IOP.Codec _codec;

	private IIOPServerProtocol _serverProtocol = null;
	private ServerManager _serverManager = null;

	private Map _channels = new WeakValueHashMap();

	private CodeSetComponentInfo _codesetInfo;

	private TransportClientInitializer _tpInit;

  private org.omg.IOP.ServiceContext _biDirSC;

	/**
	 * An orb reference.
	 */
	public org.omg.CORBA.ORB orb() {
		return _orb;
	}

	/**
	 * Get a reference to the orb's client manager.
	 */
	public ClientManager getClientManager() {
		return _clientManager;
	}

	/**
	 * Create addresses from component in IOR
	 */
	public Address [] createAddresses(org.omg.GIOP.IORAddressingInfo address) {
		return _tpInit.establishTransports(IIOPAddress.get_addresses(_codec, address.ior, address.selected_profile_index));
	}

	/**
	 * Returns a Collection of ClientBinding objects, prioritised at the per-profile
	 * (inter-component) level. The client addresses in the returned bindings
	 * should return identical results for each of the addressing disposition
	 * types.
	 */
	public ClientBinding createBinding(Address addr) {
		if(!(addr instanceof IIOPAddress))
			return null;

		IIOPAddress address = (IIOPAddress)addr;


		ClientBinding binding;
		synchronized(_channels) {
			try {
				TransportAssociationHolder tpAssc = new TransportAssociationHolder();
				IIOPClientChannel channel = findCreateChannel(address, null, tpAssc);

				if(_serverManager == null || _serverProtocol == null
					|| !_serverProtocol.servesAddress(address.get_hostname(), address.get_port()))
					binding = new ClientBinding(address, channel, tpAssc.value);
				else
					binding = new ClientBinding(address, channel, tpAssc.value, _serverManager);
			}
			catch(org.omg.CORBA.SystemException ex) {
				binding = new ClientBinding(address, ex);
			}
		}

		return binding;
	}

    /* method modified by Diana on August 23 2001*/

    	public ClientBinding createBinding(Address addr,int version) {
		if(!(addr instanceof IIOPAddress))
			return null;

		IIOPAddress address = (IIOPAddress)addr;


		ClientBinding binding;
		synchronized(_channels) {
			try {
				TransportAssociationHolder tpAssc = new TransportAssociationHolder();
				IIOPClientChannel channel = findCreateChannel(address, null, tpAssc);

				if(_serverManager == null || _serverProtocol == null
					|| !_serverProtocol.servesAddress(address.get_hostname(), address.get_port()))
					binding = new ClientBinding(address, channel, tpAssc.value,version);
				else
					binding = new ClientBinding(address, channel, tpAssc.value, _serverManager,version);
			}
			catch(org.omg.CORBA.SystemException ex) {
				binding = new ClientBinding(address, ex,version);
			}
		}

		return binding;
	}

	/**
	 * Find a channel. Lock on channels must be owned.
	 */
	protected IIOPClientChannel findCreateChannel(IIOPAddress address, IIOPClientChannel replace,
					TransportAssociationHolder tpAssc)
	{
		InetAddress host = address.get_host();

		if(host == null)
			throw new org.omg.CORBA.COMM_FAILURE("Host not found",
				IIOPMinorCodes.COMM_FAILURE_HOST_NOT_FOUND, org.omg.CORBA.CompletionStatus.COMPLETED_NO);

		LookupKey key = new LookupKey(address);
		IIOPClientChannel channel = (IIOPClientChannel)_channels.get(key);

		if(channel == null || replace == channel) {
		    if(address.get_port() == 0)
			throw new org.omg.CORBA.NO_PERMISSION("Port number published in IOR is 0.", IIOPMinorCodes.NO_PERMISSION_INVALID_PORT, org.omg.CORBA.CompletionStatus.COMPLETED_NO);

		    // select codesets
		    int tcsc, tcsw;

		    CodeSetComponentInfo ci = address.getCodesetComponentInfo();

		    if(ci == null) {
			tcsc = 0;
			tcsw = 0;
		    } else {
			tcsc = selectCodeset(_codesetInfo.ForCharData, ci.ForCharData, true);
			tcsw = selectCodeset(_codesetInfo.ForWcharData, ci.ForWcharData, false);
		    }

		    // create the transport
		    Transport transport = address.createTransport();

		    transport.establishAssociation(address, tpAssc);

		    // create the channel
		    channel = new IIOPClientChannel(this, transport, _serverProtocol, tcsc, tcsw);

		    _channels.put(new HostPortPair(host, address.get_port(), tcsc, tcsw, transport), channel);
		} else {
		    tpAssc.value = key.value;
		}

		return channel;
	}

	private static boolean checkCodeset(int tcs, CodeSetComponent comp) {
		// check for fallbacks
		if(tcs == 0x05010001 /* UTF-8 */ || tcs == 0x00010109 /* UTF-16 */)
			return true;

		// compare codesets
		if(comp.native_code_set != tcs) {
			for(int i = 0; i < comp.conversion_code_sets.length; ++i)
				if(comp.conversion_code_sets[i] == tcs)
					return true;
			return false;
		}
		return true;
	}

	/**
	 * Select a conversion code set.
	 * This algorithm is taken from 13.7.2.6 of the corba spec.
	 */
	private static int selectCodeset(CodeSetComponent local, CodeSetComponent server, boolean singleByte) {
		if(local.native_code_set == server.native_code_set)
			return local.native_code_set;
		for(int i = 0; i < local.conversion_code_sets.length; ++i)
			if(server.native_code_set == local.conversion_code_sets[i])
				return server.native_code_set;
		for(int i = 0; i < server.conversion_code_sets.length; ++i)
			if(local.native_code_set == server.conversion_code_sets[i])
				return local.native_code_set;
		for(int i = 0; i < server.conversion_code_sets.length; ++i)
			for(int j = 0; j < local.conversion_code_sets.length; ++j)
				if(server.conversion_code_sets[i] == local.conversion_code_sets[j])
					return server.conversion_code_sets[i];

		// use fallback, UTF-8 for sbcs, UTF-16 for mbcs
    int fallback = singleByte?0x05010001:0x00010109;
		if(CodeSetDatabase.compatableCodesets(local.native_code_set, server.native_code_set))
      return fallback;
    for(int i = 0; i < local.conversion_code_sets.length; ++i)
			if(CodeSetDatabase.compatableCodesets(server.native_code_set, local.conversion_code_sets[i]))
				return fallback;
		for(int i = 0; i < server.conversion_code_sets.length; ++i)
			if(CodeSetDatabase.compatableCodesets(local.native_code_set, server.conversion_code_sets[i]))
        return fallback;
		for(int i = 0; i < server.conversion_code_sets.length; ++i)
			for(int j = 0; j < local.conversion_code_sets.length; ++j)
				if(CodeSetDatabase.compatableCodesets(server.conversion_code_sets[i], local.conversion_code_sets[j]))
					return fallback;

    // no compatable codesets found.
    throw new org.omg.CORBA.CODESET_INCOMPATIBLE(0, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
	}

	/**
	 * This is called by a delegated client channel when it's peer channel closes
	 * and a new channel must be created to replace the current one.
	 */
	void rebindBidirDelegate(IIOPClientChannel deleg, IIOPAddress addr)
		throws RebindChannelException
	{
		IIOPClientChannel channel;
		TransportAssociationHolder tpAssc = new TransportAssociationHolder();
		synchronized(_channels) {
			channel = findCreateChannel(addr, deleg, tpAssc);
		}
		throw new RebindChannelException(channel, tpAssc.value);
	}

  /**
   * Get bidirectional service context, or null if no bidir is allowed
   */
  org.omg.IOP.ServiceContext getBiDirSC() {
    return _biDirSC;
  }

	/**
	 * This is called by a server channel when a BiDirService context arrives.
	 */
	IIOPClientChannel createBidirDelegate(IIOPServerChannel parent, org.omg.IIOP.ListenPoint [] listen_points) {
		IIOPClientChannel deleg;

		synchronized(_channels) {
			HostPortPair [] keys = new HostPortPair[listen_points.length];
			int j = 0;
			InetAddress addr;
			for(int i = 0; i < listen_points.length; ++i) {
				try {
					addr = InetAddress.getByName(listen_points[i].host);
				}
				catch(UnknownHostException ex) {
					continue;
				}
				// with null for the transport this acts as a lookup key.
				HostPortPair key = new HostPortPair(addr, listen_points[i].port,
				   parent.getTCSC(), parent.getTCSW(), null);

				IIOPClientChannel chan = (IIOPClientChannel)_channels.get(key);
        if(chan != null) {
          switch(chan.state()) {
          case ClientChannel.STATE_CONNECTED:
            return null;
          case ClientChannel.STATE_PAUSED:
            if(!chan.isDelegated())
              return null;
          case ClientChannel.STATE_CLOSED:
            chan = null;
            break;
          }
        }
				if(chan == null)
					keys[j++] = new HostPortPair(addr, listen_points[i].port,
				   parent.getTCSC(), parent.getTCSW(), parent._socketQueue.getTransport());
			}

			if(j == 0)
				return null;

			deleg = new IIOPClientChannel(this, parent);

			// delegate channels are not registered with the client manager
			// as their work and state management is performed via the parent
			// channel.
			for(int i = 0; i < j; ++i)
				if(keys[i] != null)
					_channels.put(keys[i], deleg);
		}

		return deleg;
	}

	private static class LookupKey extends TransportAssociationHolder {
		LookupKey(IIOPAddress addr) {
			_host = addr.get_host();
			_port = addr.get_port();
			_hash = _host.hashCode() ^ _port;

			_addr = addr;
			_csci = addr.getCodesetComponentInfo();
		}

		private int _hash;

		private InetAddress _host;
		private int _port;

		private CodeSetComponentInfo _csci;
		private IIOPAddress _addr;

		public int hashCode() {
			return _hash;
		}

		public boolean equals(Object obj) {
			if(!(obj instanceof HostPortPair))
				return false;

			HostPortPair hp2 = (HostPortPair)obj;

			if(hp2._hash != _hash || hp2._port != _port || !hp2._host.equals(_host))
				return false;

			if(_addr == null)
				return true;

			if(_csci != null && (!checkCodeset(hp2._tcsc, _csci.ForCharData)
					 || !checkCodeset(hp2._tcsw, _csci.ForWcharData)))
				return false;

			return hp2._transport.establishAssociation(_addr, this);
		}
	}

	private static class HostPortPair extends TransportAssociationHolder {
		public HostPortPair(InetAddress host, int port, int tcsc, int tcsw, Transport transport) {
			_host = host;
			_port = port;
			_hash = host.hashCode() ^ port;

			_tcsc = tcsc;
			_tcsw = tcsw;
			_transport = transport;
		}

		private int _hash;

		private InetAddress _host;
		private int _port;

		private int _tcsc = -1;
		private int _tcsw = -1;
		private Transport _transport;

		public int hashCode() {
			return _hash;
		}

		public boolean equals(Object obj) {
			if(obj instanceof LookupKey)
				return obj.equals(this);

			HostPortPair hp2 = (HostPortPair)obj;

			if(_transport != null && hp2._transport != null)
				return this == hp2;

			return hp2._hash == _hash && hp2._port == _port
					&& hp2._host.equals(_host) && _tcsc == hp2._tcsc
					&& _tcsw == hp2._tcsw;
		}
	}
}
