/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: AbstractInputStream.java,v 1.3 2001/08/13 03:47:34 wood Exp $
 *
 * Date         Author  Changes
 */

 /** Modified/extended by Diana Szentivanyi in June 2003
 * contact diasz@ida.liu.se
*/

package org.openorb.io;

import org.omg.CORBA.OctetSeqHolder;
import org.omg.CORBA.IntHolder;

import java.io.IOException;
import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;

/**
 * This class can be used as a base class for buffer input stream
 * implementations. It handles buffer management.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.3 $ $Date: 2001/08/13 03:47:34 $
 */


public abstract class AbstractInputStream
  extends org.omg.CORBA_2_3.portable.InputStream
  implements ExtendedInputStream
{
    public AbstractInputStream(StorageBuffer buf) {
	this.currbuf = buf;
	this.source = source;
    }

    public AbstractInputStream(BufferSource source) {
	this.currbuf = source.next();
	this.source = source;
    }

    /**
     * Current source buffer.
     */
    private StorageBuffer currbuf;

    /**
     * Source of fragments for this buffer
     */
    private BufferSource source = null;

    /**
     * Pass back overread value
     */
    private int overread = 0;

    /**
     * Exception thrown by all non IOStream operations if not null.
     */
    private org.omg.CORBA.SystemException cancelException = null;

    /**
     * Get codebase associated with stream. This will return the contents
     * of any TAG_JAVA_CODEBASE service context when unmarshaling, or otherwise
     * return null.
     */
    public String get_codebase() {
	return null;
    }

    /* Added by Diana on Feb. 15 2002*/
    protected byte[] its_arg_buffer(){
	if(currbuf==null) return new byte[0];
	return currbuf.its_arg_buffer();
    }


    /**
     * A read single octet from stream. This will return -1 for end of
     * stream rather than throwing an exception.
     */
    public int read()
	throws IOException
    {
	if(overread > 0)
	    return -1;
	if(cancelException != null)
	    throw cancelException;
	try {
	    return (int)read_octet();
	}
	catch(org.omg.CORBA.MARSHAL ex) {
	    if(overread > 0)
		return -1;
	    // no need to call cancel here, rethrowing the exception.
	    throw ex;
	}
    }

    /**
     * Read array from stream. This will return -1 for end of
     * stream rather than throwing an exception.
     */
    public int read(byte [] val)
    {
	return read(val, 0, val.length);
    }

    /**
     * Read array from stream. This will return -1 for end of
     * stream rather than throwing an exception.
     */
    public int read(byte [] val, int off, int len)
    {
	if(overread > 0)
	    return -1;
	if(cancelException != null)
	    throw cancelException;
	try {
	    read_octet_array(val, off, len);
	    return len;
	}
	catch(org.omg.CORBA.MARSHAL ex) {
	    if(overread > 0) {
		int ret = len - overread;
		return (ret == 0)?-1:ret;
	    }
	    // no need to call cancel here, rethrowing the exception.
	    throw ex;
	}
    }

    /**
     * throws exception if skip not possible
     */
    protected void force_skip(int len) {
	len -= skip(len);
	if(len != 0) {
	    overread = len;
	    // overreads are never reported to interceptors
	    // and completion status for overreads will be maybe
	    throw new org.omg.CORBA.MARSHAL("Buffer overread by "+len+" bytes",
					    MinorCodes.MARSHAL_BUFFER_OVERREAD, CompletionStatus.COMPLETED_MAYBE);
	}
    }

    /**
     * Notify source of cancelation of input.
     */
    protected void cancel(org.omg.CORBA.SystemException ex) {
	if(source == null) {
	    cancelException = ex;
	    throw cancelException;
	}
	source.setException(ex);
	// this will throw an exception
	try {
	    source.next();
	}
	catch(org.omg.CORBA.SystemException nex) {
	    cancelException = nex;
	    throw nex;
	}
	org.openorb.util.Trace.assertTrue(false, "Invalid state");
    }

    /**
     * Available bytes in the input buffer. This can change upwards as fragments
   * arrive.
   */
	public int available() {
		return currbuf.available() + ((source != null)?source.available():0);
	}

	public boolean markSupported() {
		return true;
	}

  public void mark(int readlimit) {
    if(source != null)
      source.mark();
    else
      currbuf.mark();
  }

  public void reset()
    throws IOException
  {
    if(source != null) {
      StorageBuffer ncurr = source.reset();
      if(ncurr == null)
        throw new IOException("No marked position to reset to");
      currbuf = ncurr;
    } else if(!currbuf.reset())
      throw new IOException("No marked position to reset to");
		overread = 0;
  }

  /**
   * Skip over bytes in the input buffer.
   */
  public long skip(long count)
	{
	    if(cancelException != null)
		throw cancelException;

	    long total = count;

	    if(source != null) {
		if(currbuf == null)
		    return 0l;
		while(count > currbuf.available()) {
		    count -= currbuf.available();
		    try {
			currbuf = source.next();
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			cancelException = ex;
			throw ex;
		    }
		    if(currbuf == null)
			return total-count;
		}

		IntHolder len = new IntHolder((int)count);
		count -= currbuf.skip(len);
		if(currbuf.available() == 0) {
		    try {
			currbuf = source.next();
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			cancelException = ex;
			throw ex;
		    }
		}
	    } else {
		IntHolder len = new IntHolder((int)count);
		count -= currbuf.skip(len);
	    }

	    return total-count;
	}

	/**
	 * Handles obtaining new buffer if old one is empty
	 */
	protected int next(OctetSeqHolder dst, IntHolder off, IntHolder len) {
	    if(cancelException != null)
		throw cancelException;

	    if(len.value == 0)
		return 0;

	    // check for an overread from a BufferSource
	    if(currbuf == null) {
		// overreads are never reported to interceptors
		// and completion status for overreads will be maybe
		overread = len.value;
		throw new org.omg.CORBA.MARSHAL("Buffer overread by "+len.value+" bytes",
						MinorCodes.MARSHAL_BUFFER_OVERREAD, CompletionStatus.COMPLETED_MAYBE);
	    }

	    int ret = currbuf.next(dst, off, len);

	    // if we have reached the end of the buffer try to get a new one.
	    if(currbuf.available() == 0) {
		if(source != null) {
		    try {
			currbuf = source.next();
		    }
		    catch(org.omg.CORBA.SystemException ex) {
			cancelException = ex;
			throw ex;
		    }
		} else if(ret < 0) {
				// report non-sourced buffer overreads.
		    overread = len.value;
		    throw new org.omg.CORBA.MARSHAL("Buffer overread by "+len.value+" bytes",
						    MinorCodes.MARSHAL_BUFFER_OVERREAD, CompletionStatus.COMPLETED_MAYBE);
		}
	    }
	    return ret;
	}
}
