/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: StorageBuffer.java,v 1.2 2001/03/22 02:37:37 wood Exp $
 *
 * Date         Author  Changes
 */

 /** Modified/extended by Diana Szentivanyi in June 2003
 * contact diasz@ida.liu.se
*/


package org.openorb.io;

import java.io.*;
import org.omg.CORBA.IntHolder;
import org.omg.CORBA.OctetSeqHolder;

/**
 * A storage buffer holds binary data for further processing. MarshalBuffers
 * and BufferSource objects output a stream of StorageBuffers.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.2 $ $Date: 2001/03/22 02:37:37 $
 */
public class StorageBuffer {

  /**
   * Public constructor, Create storage buffer from bytes. The bytes are
   * considered read-only.
   */
  public StorageBuffer(byte [] buf, int off, int len)
  {
    if(off < 0 || len < 0 || len+off > buf.length)
      throw new IndexOutOfBoundsException();

    head = new Scrap();
    avail = len;

    head.buf = buf;
    head.off = off;
    head.len = len;
    head.pos = len;
    head.mode = Scrap.MODE_READONLY;
  }

	/**
	 * public constructor, read from stream.
	 */
	public StorageBuffer(InputStream is, int total_len)
		throws IOException
  {
    this(null, 0, 0, is, total_len);
	}

  /**
   * Public constructor, read part from buffer and remainer from stream.
   * A whole buffer is always read. If the reading thread is interrupted
   * the buffer will be completley read and the interrupted status reset.
   *
	 * @param buf Prefix bytes. Copied.
	 * @param off Offset into prefix buffer.
	 * @param len Length of data in prefix buffer.
   * @param is Input stream for remainder of data.
   * @param total_len total length including data from prexix and stream.
   */
  public StorageBuffer(byte [] buf, int off, int len,
	                     InputStream is, int total_len)
		throws IOException
  {
    if(buf != null && (off < 0 || len < 0 || len+off > buf.length || (is != null && len>total_len)))
      throw new IndexOutOfBoundsException();

    Scrap t = new Scrap();
    Scrap h = t;

    boolean interrupt = false;

    this.avail = total_len;
    if(buf == null || len == 0) {
      t.pos = 0;
      t.buf = null;
    } else if(is == null) {
      t.buf = new byte[len];
      t.off = 0;
      t.len = len;
      t.pos = len;
      t.mode = Scrap.MODE_NORMAL;

      System.arraycopy(buf, off, t.buf, 0, len);

      head = h;
      this.avail = len;
      return;
    } else {
      int alloc = (total_len > Scrap.DEFAULT_SCRAP)?Scrap.DEFAULT_SCRAP:total_len;
      t.buf = new byte[alloc];
      t.off = 0;
      t.len = alloc;
      t.pos = alloc;
      t.mode = Scrap.MODE_NORMAL;

      System.arraycopy(buf, off, t.buf, 0, len);
      alloc -= len;
      total_len -= len;
      int pos = len;
      while(alloc > 0) {
        int ret;
        try {
          ret = is.read(t.buf, pos, alloc);
        }
        catch(InterruptedIOException ex) {
          interrupt = true;
          ret = ex.bytesTransferred;
        }
        if(ret >= 0) {
          pos += ret;
          alloc -= ret;
          total_len -= ret;
        } else {
          if(interrupt)
            Thread.currentThread().interrupt();
          throw new EOFException("error: EOF reached when reading message");
        }
      }
    }

    while(total_len > 0) {
      int alloc = (total_len > Scrap.DEFAULT_SCRAP)?Scrap.DEFAULT_SCRAP:total_len;
      Scrap nex = new Scrap();
      nex.buf = new byte[alloc];
      nex.off = 0;
      nex.len = alloc;
      nex.pos = t.pos + alloc;
      nex.mode = Scrap.MODE_NORMAL;

      int pos = 0;
      while(alloc > 0) {
        int ret;
        try {
          ret = is.read(nex.buf, pos, alloc);
        }
        catch(InterruptedIOException ex) {
          interrupt = true;
          ret = ex.bytesTransferred;
        }
        if(ret >= 0) {
          pos += ret;
          alloc -= ret;
          total_len -= ret;
        } else {
          if(interrupt)
            Thread.currentThread().interrupt();
          throw new EOFException("error: EOF reached when reading message");
        }
      }

      t.next = nex;
      t = nex;
    }
    if(interrupt)
      Thread.currentThread().interrupt();

    if(h.buf == null)
      head = h.next;
    else
      head = h;
  }

	/**
	 * package protected constructor. Used by MarshallBuffer.
	 */
	StorageBuffer(Scrap head, int len) {
		this.head = head;
		this.tail = null;
		this.avail = len;
	}

	/**
	 * package protected constructor. Used by MarshallBuffer when
	 * creating preview buffers.
	 */
	StorageBuffer(Scrap head, Scrap tail) {
		this.head = head;
		this.tail = tail;
		this.avail = tail.pos - tail.len - head.pos + head.len;
	}

    /* Added by Diana on feb. 15 2002*/

    public byte[] its_arg_buffer(){
	byte[] ret=new byte[avail];
	System.arraycopy(head.buf,head.buf.length-avail,ret,0,avail);
	return ret;
    }

  /**
   * Available bytes
   */
	public int available() {
		return avail;
	}

	/**
	 * Writes entire buffer to output stream.
	 */
	public void writeTo(OutputStream os)
		throws IOException
  {
		if(head == null)
			throw new EOFException("error: buffer is empty");

    boolean interrupt = Thread.interrupted();

		if(readWriteMode)
			while(head != tail) {
				if(head.mode == Scrap.MODE_READONLY) {
          byte [] tmp = new byte[head.len];
					System.arraycopy(head.buf, head.off, tmp, 0, head.len);
          head.buf = tmp;
          head.off = 0;
          head.mode = Scrap.MODE_NORMAL;
				}
        int d = 0;
        while(d < head.len) {
          try {
            os.write(head.buf, head.off+d, head.len-d);
            break;
          }
          catch(InterruptedIOException ex) {
            interrupt = true;
            d += ex.bytesTransferred;
          }
        }
				head = head.next;
			}
		else
			while(head != tail) {
        int d = 0;
        while(d < head.len) {
          try {
    				os.write(head.buf, head.off+d, head.len-d);
            break;
          }
          catch(InterruptedIOException ex) {
            interrupt = true;
            d += ex.bytesTransferred;
          }
        }
				head = head.next;
			}

		head = null;
		avail = 0;

    if(interrupt)
      Thread.currentThread().interrupt();
	}

	/**
	 * Read the next piece of the buffer. This buffer is modifyable if
	 * in readWriteMode.
	 *
	 * @param buf Out parameter, holds pointer to scratch space on return.
	 * @param off Out parameter, holds buffer offset on return.
	 * @param len InOut parameter, Length of requested buffer. Holds
	 * unallocated length on return. (ie: original - return)
	 *
	 * @return Length of returned buf. If end of buffer has been reached
	 * -1 is returned.
	 */
	public int next(OctetSeqHolder buf, IntHolder off, IntHolder len) {
	    if(head == null) {
		buf.value = null;
		off.value = 0;
		return -1;
	    }

	    if(len.value < 0)
		throw new IndexOutOfBoundsException();

	    if(readWriteMode && head.mode == Scrap.MODE_READONLY) {
		byte [] copy = new byte[head.len];
		System.arraycopy(head.buf, head.off, copy, 0, head.len);
		head.buf = copy;
		head.off = 0;
		head.mode = Scrap.MODE_NORMAL;
	    }

	    buf.value = head.buf;
	    off.value = head.off;
	    if(len.value < head.len) {
		int olen = len.value;
		len.value = 0;
		if((mark == null && tail == null) || head == temphead) {
		    head.off += olen;
		    head.len -= olen;
		} else {
				// the real head may be used again later, use a tempory head
		    temphead.buf = head.buf;
		    temphead.off = head.off + olen;
		    temphead.len = head.len - olen;
		    temphead.mode = head.mode | Scrap.MODE_SHARED;
		    temphead.pos = head.pos;
		    temphead.next = head.next;
		    head = temphead;
		}
		avail -= olen;
		return olen;
	    } else {
		int olen = head.len;
		len.value -= olen;
		if((head = head.next) == tail)
		    head = null;
		avail -= olen;
		return olen;
	    }
	}

  /**
   * Skip over bytes.
   *
	 * @param len InOut parameter, Length of requested skip. Holds
	 * unskipped length on return. (ie: original - return)
	 *
	 * @return Length of returned buf. If end of buffer has been reached
	 * -1 is returned.
   */
	public int skip(IntHolder len) {
		if(head == null)
			return -1;

		int skip = 0;

		while(head != null && len.value >= head.len) {
			len.value -= head.len;
			avail -= head.len;
			skip += head.len;
			if((head = head.next) == tail)
        head = null;
		}

		if(head != null && len.value > 0) {
			if((mark == null && tail == null) || head == temphead) {
				head.off += len.value;
				head.len -= len.value;
			} else {
				// the real head may be used again later, use a tempory head
				temphead.buf = head.buf;
				temphead.off = head.off + len.value;
				temphead.len = head.len - len.value;
				temphead.mode = head.mode | Scrap.MODE_SHARED;
				temphead.pos = head.pos;
				temphead.next = head.next;
				head = temphead;
			}
			skip += len.value;
			avail -= len.value;
			len.value = 0;
		}
		return skip;
	}

  /**
   * Move the entire buffer into a single byte array. Calls to this function
	 * should generaly be avoided, call mark and next to iterate over the buffer
	 * instead.
   *
   * @returns the linearized buffer.
   */
  public byte [] linearize()
  {
    if(head == null)
      return new byte[0];

    if(head.len != head.buf.length || head.next != tail || head.mode != Scrap.MODE_NORMAL || head.off != 0) {
      byte [] buf = new byte[avail];

      Scrap nex = head;
      int disp = head.pos-head.len;
      while(nex != tail) {
        System.arraycopy(nex.buf, nex.off, buf, disp, nex.len);
        disp += nex.len;
        nex = nex.next;
      }
      head.buf = buf;
      head.off = 0;
      head.len = avail;
      head.mode = Scrap.MODE_NORMAL;
      head.pos = head.pos-head.len + avail;
      head.next = tail;
    }

    return head.buf;
  }

	/**
	 * Check if the buffer is in read write mode. In read write mode buffers
	 * returned from the next operation can be safely modified, modifying the
	 * contents of the buffer.
	 */
	public boolean isReadWriteMode() {
		return readWriteMode;
	}

	/**
	 * Set read write mode. In read write mode buffers
	 * returned from the next operation can be safely modified, modifying the
	 * contents of the buffer.
	 */
	public void setReadWriteMode(boolean readWriteMode) {
		this.readWriteMode = readWriteMode;
	}

	/**
	 * Mark current buffer position for future reset.
	 */
	public boolean mark() {
		if(mark == null) {
			mark = head;
			markavail = avail;
			return true;
		}
		return false;
	}

	/**
	 * reset buffer position to position marked with mark operation.
	 */
	public boolean reset() {
		if(mark != null) {
			head = mark;
			avail = markavail;
			mark = null;
			return true;
		}
		return false;
	}

	private int avail;

	private Scrap mark = null;
	private int markavail = 0;

	private Scrap head;
	private Scrap temphead = new Scrap();
	private Scrap tail = null;

	private boolean readWriteMode = false;
}
