/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: ServerRequest.java,v 1.4 2001/07/31 13:45:34 wood Exp $
 *
 * Date         Author  Changes
 */


/** Modified/extended by Diana Szentivanyi in June 2003
* contact diasz@ida.liu.se
*/

package org.openorb.net;

import org.openorb.adapter.*;

/**
 * Interface for server requests. The server request holds all the data
 * concering a request. It is responsible for interpreting incoming requests,
 * locating an adapter containing the target servant, calling interception
 * points and construting a reply.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.4 $ $Date: 2001/07/31 13:45:34 $
 */
public interface ServerRequest
	extends org.omg.PortableInterceptor.ServerRequestInfo,
      org.omg.CORBA.portable.ResponseHandler
{
  /**
   * ORB reference.
   */
  public org.omg.CORBA.ORB orb();

	/**
	 * Server channel
	 */
	public ServerChannel channel();

	/**
	 * Current request state.
	 */
	public int state();

  /**
   * Get the object_key for the request. This is always available.
   */
  public byte [] object_key();

  /**
   * Get all request service contexts that have been set. Valid in REPLY and
   * COMPLETE.
   */
  public org.omg.IOP.ServiceContext [] get_reply_service_contexts();

  /**
   * This returns true if this request is a locate request. Use replyLocate
   * to set the reply.
   *
   * Interceptors are not called for locate requests.
   *
   * ServerRequestInfo operations: arguments, exceptions, contexts,
   * operation_context, result, get_request_service_context,
   * get_reply_service_context, add_reply_service_context are not valid.
   *
   * operation returns the empty string "", response_expected returns true.
   *
   * Note the extra value for reply_status, UNKNOWN_OBJECT.
   */
  public boolean is_locate();

	/**
	 * Get the transport association.
	 */
	public Object get_transport_association();

	/**
   * Extra reply status denoting a locate request found the object.
   */
  public final static short OBJECT_HERE = -1;

  /**
   * Extra reply status denoting a locate request where the object is not found
   */
  public final static short UNKNOWN_OBJECT = -2;

  /**
   * The request has arrived but has not had any of it's service contexts
   * examined. A cancelation in this state will result in COMPLETED_NO status.
   * Calls valid in receive_request_service_contexts are valid, this
   * interception point will occour while in this state if the request is
   * not a locate request.
   */
  public static final int STATE_CREATED = 0;

  /**
   * Client intiated request cancelation. This may be called at any time to
   * indicate that a response from the request is no longer expected.
   */
  public void client_cancel();

  /**
   * Server intiated request cancelation. This may be called at any time apart
   * from during a find_adapter or dispatch call to indicate a server
   * initiated cancel. This will send a reply to the client.
   */
  public void server_cancel(Throwable ex);

  /**
   * Called by the worker thread before every attempt to run the request. This
   * should call the recieve_request_service_context interception point.
   * @returns the state of the request.
   */
  public int begin_request();

  /**
   * Set the object adapter and transfer to the QUEUED state. This is legal
   * only in the FIND_ADAPTER or COMPLETE states. If COMPLETE this has no effect.
   *
   * @returns the state of the request. This will be either QUEUED or COMPLETE.
   */
  public int adapter(ObjectAdapter adapter);

  /**
	 * Server adapter. This operation is valid in all states apart from
	 * CREATED. When a dispatch operation fails this will return the original
   * adapter.
	 */
	public ObjectAdapter adapter();

	/**
	 * This state is entered if a parent adapter is either holding requests
	 * or is in the process of being destroyed. To continue find_adapter must
	 * be called again.
	 */
	public static final int STATE_FIND_ADAPTER = 1;

  /**
   * Request has had an adapter found and is enqueued awaiting dispatch.
   * Calls valid in receive_request_service_contexts are valid. Request
   * fragments may still be arriving.
   */
  public static final int STATE_QUEUED = 2;

  /**
   * Dispatch a request to it's destination. This is valid in the QUEUED state
   * Call predispatch on the adapter to locate the target, call the
   * receive_request interception points, enter the PROCESSING state and
   * dispatch the request through the adapter. If sync scope is
   * SYNC_WITH_SERVER an empty response will be sent before the request
   * is dispatched. When this function returns all the terminating
   * interception points will have been called, and state will be COMPLETE.
   * If the target adapter is destroyed between the find_adapter and dispatch
   * this function will return false and the CREATED state is re-entered, in
   * this case the find_adapter function will be called again once the
   * unregister adapter operation is called on the server manager with the
   * target adapter.
   */
  public void dispatch()
    throws AdapterDestroyedException;

  /**
   * Request is being processed. This is the only state where the
   * ReplyHandler functions can be successfully called. Request
   * fragments may still be arriving.
   */
  public static final int STATE_PROCESSING = 3;

  /**
   * Get the request input stream. This function can be called only once
   * while in the PROCESSING state. The returned input stream may throw a
   * system exception if some problem occours while unmarshalling the request
   * arguments. This function, and the reply functions must be called by the
   * same thread, the thread that called the dispatch function.
   */
  public org.omg.CORBA.portable.InputStream argument_stream();

    /* Modified by Diana on Oct. 24 2001*/

    public void argument_stream(org.omg.CORBA.portable.InputStream is);


  // response handler interface.

  /**
   * Create a stream for marshaling the reply. The send_reply interception
   * points are called, if this results in a system exception the exception
   * is thrown, otherwise a stream is retured. The returned stream may raise
   * a system exception when attempting to write to it indicating a
   * transmission problem. Such an exception is not reported to interceptors.
   * This function is inherited from the ResponseHandler interface.
   */
  public org.omg.CORBA.portable.OutputStream createReply();


    /* Method introduced by Diana on Aug. 29 2001*/
    public org.omg.CORBA.portable.OutputStream createNoInvokeReply();

  /**
   * Create a stream for marshaling a user exception. The send_exception
   * interception points are called, if this results in a system exception the
   * exception is thrown, otherwise a stream is retured. The returned stream
   * may raise a system exception when attempting to write to it indicating a
   * transmission problem. Such an exception is not reported to interceptors.
   * This function is inherited from the ResponseHandler interface.
   */
  public org.omg.CORBA.portable.OutputStream createExceptionReply();

  /**
   * Replying, one of the response handlers functions have been called or an
   * an internal response is being constructed. send_* interception points are
   * called in this state.
   */
  public static final int STATE_REPLY = 4;

  /**
   * Request has completed. Terminating state.
   */
  public static final int STATE_COMPLETE = 5;

  /**
   * Get the ior associated with a forward reference. Calling this function
   * instead of forward_reference avoids creating an enclosing delegate/object.
   * This function is valid in REPLY and COMPLETE states where reply_status
   * would return LOCATION_FORWARD or LOCATION_FORWARD_PERMANENT.
   */
  public org.omg.IOP.IOR forward_reference_ior();

  /**
   * Get the system exception which would be contained in the any returned
   * from sending_exception. This function is valid in REPLY and COMPLETE
   * states where reply_status would return SYSTEM_EXCEPTION.
   */
  public org.omg.CORBA.SystemException sending_system_exception();

  /**
   * Get the ID of the system exception which would be contained in the
   * any returned from sending_exception. This function is valid in REPLY
   * and COMPLETE states where reply_status would return SYSTEM_EXCEPTION.
   */
  public String sending_system_exception_id();

    public void result(org.omg.CORBA.Any r);

    public org.omg.CORBA.Any result();

    public org.omg.CORBA.Any[] result_arguments();

    public void result_arguments(org.omg.CORBA.Any[] ra);

    public void client_id_(String cl_id);

    public void retention_id_(int ret_id);
    public void expiration_time_(long exp_time);

    public int _retention_id();

    public String _client_id();
    public long _expiration_time();
}
